from abc import ABC, abstractmethod
import asyncio

try:
    from datetime import UTC, datetime
except ImportError:
    from datetime import datetime, timezone

    UTC = timezone.utc

from struct import unpack
from typing import Generic
from typing_extensions import Self, TypeVar

from amqtt.adapters import ReaderAdapter, WriterAdapter
from amqtt.codecs_amqtt import bytes_to_hex_str, decode_packet_id, int_to_bytes, read_or_raise
from amqtt.errors import CodecError, MQTTError, NoDataError

RESERVED_0 = 0x00
CONNECT = 0x01
CONNACK = 0x02
PUBLISH = 0x03
PUBACK = 0x04
PUBREC = 0x05
PUBREL = 0x06
PUBCOMP = 0x07
SUBSCRIBE = 0x08
SUBACK = 0x09
UNSUBSCRIBE = 0x0A
UNSUBACK = 0x0B
PINGREQ = 0x0C
PINGRESP = 0x0D
DISCONNECT = 0x0E
RESERVED_15 = 0x0F


class MQTTFixedHeader:
    """Represents the fixed header of an MQTT packet."""

    __slots__ = ("flags", "packet_type", "remaining_length")

    def __init__(self, packet_type: int, flags: int = 0, length: int = 0) -> None:
        self.packet_type = packet_type
        self.flags = flags
        self.remaining_length = length

    def to_bytes(self) -> bytes:
        """Encode the fixed header to bytes."""

        def encode_remaining_length(length: int) -> bytes:
            """Encode the remaining length as per MQTT protocol."""
            encoded = bytearray()
            while True:
                length_byte = length % 0x80
                length //= 0x80
                if length > 0:
                    length_byte |= 0x80
                encoded.append(length_byte)
                if length <= 0:
                    break
            return bytes(encoded)

        try:
            packet_type_flags = (self.packet_type << 4) | self.flags
            encoded_length = encode_remaining_length(self.remaining_length)
            return bytes([packet_type_flags]) + encoded_length
        except OverflowError as exc:
            msg = f"Fixed header encoding failed: {exc}"
            raise CodecError(msg) from exc

    async def to_stream(self, writer: WriterAdapter) -> None:
        """Write the fixed header to the stream."""
        writer.write(self.to_bytes())

    @property
    def bytes_length(self) -> int:
        return len(self.to_bytes())

    @classmethod
    async def from_stream(cls: type[Self], reader: ReaderAdapter) -> "Self | None":
        """Decode a fixed header from the stream."""

        async def decode_remaining_length() -> int:
            """Decode the remaining length from the stream."""
            multiplier: int
            value: int
            multiplier, value = 1, 0
            buffer = bytearray()
            while True:
                encoded_byte = await reader.read(1)
                byte_value = unpack("!B", encoded_byte)[0]
                buffer.append(byte_value)
                value += (byte_value & 0x7F) * multiplier
                if (byte_value & 0x80) == 0:
                    break
                multiplier *= 128
                if multiplier > 128**3:
                    msg = f"Invalid remaining length bytes:{bytes_to_hex_str(buffer)}, packet_type={packet_type}"
                    raise MQTTError(msg)
            return value

        try:
            byte1 = await read_or_raise(reader, 1)
            int1 = unpack("!B", byte1)[0]
            packet_type = (int1 & 0xF0) >> 4
            flags = int1 & 0x0F
            remaining_length = await decode_remaining_length()
            return cls(packet_type, flags, remaining_length)
        except NoDataError:
            return None

    def __repr__(self) -> str:
        """Return a string representation of the MQTTFixedHeader object."""
        return f"{self.__class__.__name__}(packet_type={self.packet_type}, flags={self.flags}, length={self.remaining_length})"


_FH = TypeVar("_FH", bound=MQTTFixedHeader)


class MQTTVariableHeader(ABC):
    """Abstract base class for MQTT variable headers."""

    async def to_stream(self, writer: asyncio.StreamWriter) -> None:
        writer.write(self.to_bytes())
        await writer.drain()

    @abstractmethod
    def to_bytes(self) -> bytes | bytearray:
        """Serialize the variable header to bytes."""

    @property
    def bytes_length(self) -> int:
        return len(self.to_bytes())

    @classmethod
    @abstractmethod
    async def from_stream(cls: type[Self], reader: ReaderAdapter, fixed_header: MQTTFixedHeader) -> Self:
        pass


class PacketIdVariableHeader(MQTTVariableHeader):
    """Represents a variable header containing a packet ID."""

    __slots__ = ("packet_id",)

    def __init__(self, packet_id: int) -> None:
        super().__init__()
        self.packet_id = packet_id

    def to_bytes(self) -> bytes:
        return int_to_bytes(self.packet_id, 2)

    @classmethod
    async def from_stream(
        cls: type[Self],
        reader: ReaderAdapter,
        _: MQTTFixedHeader | None = None,
    ) -> Self:
        packet_id = await decode_packet_id(reader)
        return cls(packet_id)

    def __repr__(self) -> str:
        """Return a string representation of the PacketIdVariableHeader object."""
        return f"{self.__class__.__name__}(packet_id={self.packet_id})"


_VH = TypeVar("_VH", bound=MQTTVariableHeader | None)


class MQTTPayload(Generic[_VH], ABC):
    """Abstract base class for MQTT payloads."""

    async def to_stream(self, writer: asyncio.StreamWriter) -> None:
        writer.write(self.to_bytes())
        await writer.drain()

    @abstractmethod
    def to_bytes(self, fixed_header: MQTTFixedHeader | None = None, variable_header: _VH | None = None) -> bytes | bytearray:
        pass

    @classmethod
    @abstractmethod
    async def from_stream(
        cls: type[Self],
        reader: asyncio.StreamReader | ReaderAdapter,
        fixed_header: MQTTFixedHeader | None,
        variable_header: _VH | None,
    ) -> Self:
        pass


_P = TypeVar("_P", bound=MQTTPayload[MQTTVariableHeader] | None)


class MQTTPacket(Generic[_VH, _P, _FH]):
    """Represents an MQTT packet."""

    __slots__ = ("fixed_header", "payload", "protocol_ts", "variable_header")

    VARIABLE_HEADER: type[_VH] | None = None
    PAYLOAD: type[_P] | None = None
    FIXED_HEADER: type[_FH] = MQTTFixedHeader  # type: ignore [assignment]

    def __init__(self, fixed: _FH, variable_header: _VH | None = None, payload: _P | None = None) -> None:
        self.fixed_header = fixed
        self.variable_header = variable_header
        self.payload = payload
        self.protocol_ts: datetime | None = None

    async def to_stream(self, writer: WriterAdapter) -> None:
        """Write the entire packet to the stream."""
        writer.write(self.to_bytes())
        await writer.drain()
        self.protocol_ts = datetime.now(UTC)

    def to_bytes(self) -> bytes:
        """Serialize the packet into bytes."""
        variable_header_bytes = self.variable_header.to_bytes() if self.variable_header is not None else b""
        payload_bytes = self.payload.to_bytes(self.fixed_header, self.variable_header) if self.payload is not None else b""

        fixed_header_bytes = b""
        if self.fixed_header:
            self.fixed_header.remaining_length = len(variable_header_bytes) + len(payload_bytes)
            fixed_header_bytes = self.fixed_header.to_bytes()

        return fixed_header_bytes + variable_header_bytes + payload_bytes

    @classmethod
    async def from_stream(
        cls: type[Self],
        reader: ReaderAdapter,
        fixed_header: _FH | None = None,
        variable_header: _VH | None = None,
    ) -> Self:
        """Decode an MQTT packet from the stream."""
        if fixed_header is None:
            fixed_header = await cls.FIXED_HEADER.from_stream(reader)

        if cls.VARIABLE_HEADER and variable_header is None:
            variable_header = await cls.VARIABLE_HEADER.from_stream(reader, fixed_header)

        if cls.PAYLOAD and fixed_header:
            payload = await cls.PAYLOAD.from_stream(reader, fixed_header, variable_header)
        else:
            payload = None

        if fixed_header and not variable_header and not payload:
            instance = cls(fixed_header)
        elif fixed_header and not payload:
            instance = cls(fixed_header, variable_header)
        else:
            instance = cls(fixed_header, variable_header, payload)
        instance.protocol_ts = datetime.now(UTC)
        return instance

    @property
    def bytes_length(self) -> int:
        return len(self.to_bytes())

    def __repr__(self) -> str:
        """Return a string representation of the packet."""
        return (
            f"{self.__class__.__name__}(ts={self.protocol_ts}, "
            f"fixed={self.fixed_header}, variable={self.variable_header}, payload={self.payload})"
        )



from typing_extensions import Self

from amqtt.adapters import ReaderAdapter
from amqtt.codecs_amqtt import bytes_to_int, read_or_raise
from amqtt.errors import AMQTTError
from amqtt.mqtt.packet import CONNACK, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader

CONNECTION_ACCEPTED = 0x00
UNACCEPTABLE_PROTOCOL_VERSION = 0x01
IDENTIFIER_REJECTED = 0x02
SERVER_UNAVAILABLE = 0x03
BAD_USERNAME_PASSWORD = 0x04
NOT_AUTHORIZED = 0x05


class ConnackVariableHeader(MQTTVariableHeader):
    __slots__ = ("return_code", "session_parent")

    def __init__(self, session_parent: int | None = None, return_code: int | None = None) -> None:
        super().__init__()
        self.session_parent = session_parent
        self.return_code = return_code

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, _: MQTTFixedHeader | None) -> Self:
        data = await read_or_raise(reader, 2)
        session_parent = data[0] & 0x01
        return_code = bytes_to_int(data[1])
        return cls(session_parent, return_code)

    def to_bytes(self) -> bytes | bytearray:
        out = bytearray(2)
        # Connect acknowledge flags
        out[0] = 1 if self.session_parent else 0
        # Return code
        out[1] = self.return_code or 0
        return out

    def __repr__(self) -> str:
        """Return a string representation of the ConnackVariableHeader object."""
        return f"{type(self).__name__}(session_parent={hex(self.session_parent or 0)}, return_code={hex(self.return_code or 0)})"


class ConnackPacket(MQTTPacket[ConnackVariableHeader, MQTTPayload[MQTTVariableHeader], MQTTFixedHeader]):
    VARIABLE_HEADER = ConnackVariableHeader
    PAYLOAD = MQTTPayload[MQTTVariableHeader]

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: ConnackVariableHeader | None = None,
        payload: MQTTPayload[MQTTVariableHeader] | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(CONNACK, 0x00)
        elif fixed.packet_type != CONNACK:
            msg = f"Invalid fixed packet type {fixed.packet_type} for ConnackPacket init"
            raise AMQTTError(msg) from None
        else:
            header = fixed

        super().__init__(header, variable_header, payload)

    @classmethod
    def build(cls, session_parent: int | None = None, return_code: int | None = None) -> Self:
        v_header = ConnackVariableHeader(session_parent, return_code)
        return cls(variable_header=v_header)

    @property
    def return_code(self) -> int | None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.return_code

    @return_code.setter
    def return_code(self, return_code: int | None) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.return_code = return_code

    @property
    def session_parent(self) -> int | None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.session_parent

    @session_parent.setter
    def session_parent(self, session_parent: int | None) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.session_parent = session_parent



import asyncio
from collections import deque  # pylint: disable=C0412
from dataclasses import dataclass
from typing import Any, SupportsIndex, SupportsInt, TypeAlias  # pylint: disable=C0412

import psutil

from amqtt.plugins.base import BasePlugin
from amqtt.session import Session

try:
    from collections.abc import Buffer
except ImportError:
    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class Buffer(Protocol):  #  type: ignore[no-redef]
        def __buffer__(self, flags: int = ...) -> memoryview:
            """Mimic the behavior of `collections.abc.Buffer` for python 3.10-3.12."""


try:
    from datetime import UTC, datetime
except ImportError:
    from datetime import datetime, timezone

    UTC = timezone.utc


import amqtt
from amqtt.broker import BrokerContext
from amqtt.codecs_amqtt import float_to_bytes_str, int_to_bytes_str
from amqtt.mqtt.packet import PUBLISH, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader

DOLLAR_SYS_ROOT = "$SYS/broker/"
STAT_BYTES_SENT = "bytes_sent"
STAT_BYTES_RECEIVED = "bytes_received"
STAT_MSG_SENT = "messages_sent"
STAT_MSG_RECEIVED = "messages_received"
STAT_PUBLISH_SENT = "publish_sent"
STAT_PUBLISH_RECEIVED = "publish_received"
STAT_START_TIME = "start_time"
STAT_CLIENTS_MAXIMUM = "clients_maximum"
STAT_CLIENTS_CONNECTED = "clients_connected"
STAT_CLIENTS_DISCONNECTED = "clients_disconnected"
MEMORY_USAGE_MAXIMUM = "memory_maximum"
CPU_USAGE_MAXIMUM = "cpu_usage_maximum"
CPU_USAGE_LAST = "cpu_usage_last"


PACKET: TypeAlias = MQTTPacket[MQTTVariableHeader, MQTTPayload[MQTTVariableHeader], MQTTFixedHeader]


def val_to_bytes_str(value: Any) -> bytes:
    """Convert an int, float or string to byte string."""
    match value:
        case int():
            return int_to_bytes_str(value)
        case float():
            return float_to_bytes_str(value)
        case str():
            return value.encode("utf-8")
        case _:
            msg = f"Unsupported type {type(value)}"
            raise NotImplementedError(msg)


class BrokerSysPlugin(BasePlugin[BrokerContext]):
    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        # Broker statistics initialization
        self._stats: dict[str, int] = {}
        self._sys_handle: asyncio.Handle | None = None

        self._sys_interval: int = 0
        self._current_process = psutil.Process()


    def _clear_stats(self) -> None:
        """Initialize broker statistics data structures."""
        for stat in (
            STAT_BYTES_RECEIVED,
            STAT_BYTES_SENT,
            STAT_MSG_RECEIVED,
            STAT_MSG_SENT,
            STAT_CLIENTS_MAXIMUM,
            STAT_CLIENTS_CONNECTED,
            STAT_CLIENTS_DISCONNECTED,
            STAT_PUBLISH_RECEIVED,
            STAT_PUBLISH_SENT,
            MEMORY_USAGE_MAXIMUM,
            CPU_USAGE_MAXIMUM
        ):
            self._stats[stat] = 0

    async def _broadcast_sys_topic(self, topic_basename: str, data: bytes) -> None:
        """Broadcast a system topic."""
        await self.context.broadcast_message(topic_basename, data)

    def schedule_broadcast_sys_topic(self, topic_basename: str, data: bytes) -> asyncio.Task[None]:
        """Schedule broadcasting of system topics."""
        return asyncio.ensure_future(
            self._broadcast_sys_topic(DOLLAR_SYS_ROOT + topic_basename, data),
            loop=self.context.loop,
        )

    async def on_broker_pre_start(self) -> None:
        """Clear statistics before broker start."""
        self._clear_stats()

    async def on_broker_post_start(self) -> None:
        """Initialize statistics and start $SYS broadcasting."""
        self._stats[STAT_START_TIME] = int(datetime.now(tz=UTC).timestamp())
        version = f"aMQTT version {amqtt.__version__}"
        self.context.retain_message(DOLLAR_SYS_ROOT + "version", version.encode())

        # Start $SYS topics management
        try:
            self._sys_interval = self._get_config_option("sys_interval", None)
            if isinstance(self._sys_interval, str | Buffer | SupportsInt | SupportsIndex):
                self._sys_interval = int(self._sys_interval)

            if self._sys_interval > 0:
                self.context.logger.debug(f"Setup $SYS broadcasting every {self._sys_interval} seconds")
                self._sys_handle = (
                    self.context.loop.call_later(self._sys_interval, self.broadcast_dollar_sys_topics)
                    if self.context.loop is not None
                    else None
                )
            else:
                self.context.logger.debug("$SYS disabled")
        except KeyError:
            self.context.logger.debug("could not find 'sys_interval' key: {e!r}")
            # 'sys_interval' config parameter not found

    async def on_broker_pre_shutdown(self) -> None:
        """Stop $SYS topics broadcasting."""
        if self._sys_handle:
            self._sys_handle.cancel()

    def broadcast_dollar_sys_topics(self) -> None:
        """Broadcast dynamic $SYS topics updates and reschedule next execution."""
        # Update stats
        uptime = int(datetime.now(tz=UTC).timestamp()) - self._stats[STAT_START_TIME]
        client_connected = self._stats[STAT_CLIENTS_CONNECTED]
        client_disconnected = self._stats[STAT_CLIENTS_DISCONNECTED]
        inflight_in = 0
        inflight_out = 0
        messages_stored = 0
        for session in self.context.sessions:
            inflight_in += session.inflight_in_count
            inflight_out += session.inflight_out_count
            messages_stored += session.retained_messages_count
        messages_stored += len(self.context.retained_messages)
        subscriptions_count = sum(len(sub) for sub in self.context.subscriptions.values())
        self._stats[STAT_CLIENTS_MAXIMUM] = client_connected

        cpu_usage = self._current_process.cpu_percent(interval=0)
        self._stats[CPU_USAGE_MAXIMUM] = max(self._stats[CPU_USAGE_MAXIMUM], cpu_usage)

        mem_info_usage = self._current_process.memory_full_info()
        mem_size = mem_info_usage.rss / (1024 ** 2)
        self._stats[MEMORY_USAGE_MAXIMUM] = max(self._stats[MEMORY_USAGE_MAXIMUM], mem_size)

        # Broadcast updates
        tasks: deque[asyncio.Task[None]] = deque()
        stats: dict[str, int | str] = {
            "load/bytes/received": self._stats[STAT_BYTES_RECEIVED],
            "load/bytes/sent": self._stats[STAT_BYTES_SENT],
            "messages/received": self._stats[STAT_MSG_RECEIVED],
            "messages/sent": self._stats[STAT_MSG_SENT],
            "time": int(datetime.now(tz=UTC).timestamp()),
            "uptime": str(uptime),
            "uptime/formatted": str(datetime.fromtimestamp(self._stats[STAT_START_TIME], UTC)),
            "clients/connected": client_connected,
            "clients/disconnected": client_disconnected,
            "clients/maximum": self._stats[STAT_CLIENTS_MAXIMUM],
            "clients/total": client_connected + client_disconnected,
            "messages/inflight": inflight_in + inflight_out,
            "messages/inflight/in": inflight_in,
            "messages/inflight/out": inflight_out,
            "messages/inflight/stored": messages_stored,
            "messages/publish/received": self._stats[STAT_PUBLISH_RECEIVED],
            "messages/publish/sent": self._stats[STAT_PUBLISH_SENT],
            "messages/retained/count": len(self.context.retained_messages),
            "messages/subscriptions/count": subscriptions_count,
            "heap/size": mem_size,
            "heap/maximum": self._stats[MEMORY_USAGE_MAXIMUM],
            "cpu/percent": cpu_usage,
            "cpu/maximum": self._stats[CPU_USAGE_MAXIMUM],
        }
        for stat_name, stat_value in stats.items():
            data: bytes = val_to_bytes_str(stat_value)
            tasks.append(self.schedule_broadcast_sys_topic(stat_name, data))

        # Wait until broadcasting tasks end
        while tasks and tasks[0].done():
            tasks.popleft()

        # Reschedule
        self.context.logger.debug(f"Broadcast $SYS topics again in {self._sys_interval} seconds.")
        self._sys_handle = (
            self.context.loop.call_later(self._sys_interval, self.broadcast_dollar_sys_topics)
            if self.context.loop is not None
            else None
        )

    async def on_mqtt_packet_received(self, *, packet: PACKET, session: Session | None = None) -> None:
        """Handle incoming MQTT packets."""
        if packet:
            packet_size = packet.bytes_length
            self._stats[STAT_BYTES_RECEIVED] += packet_size
            self._stats[STAT_MSG_RECEIVED] += 1
            if packet.fixed_header.packet_type == PUBLISH:
                self._stats[STAT_PUBLISH_RECEIVED] += 1

    async def on_mqtt_packet_sent(self, *, packet: PACKET, session: Session | None = None) -> None:
        """Handle sent MQTT packets."""
        if packet:
            packet_size = packet.bytes_length
            self._stats[STAT_BYTES_SENT] += packet_size
            self._stats[STAT_MSG_SENT] += 1
            if packet.fixed_header.packet_type == PUBLISH:
                self._stats[STAT_PUBLISH_SENT] += 1

    async def on_broker_client_connected(self, client_id: str, client_session: Session) -> None:
        """Handle broker client connection."""
        self._stats[STAT_CLIENTS_CONNECTED] += 1
        self._stats[STAT_CLIENTS_MAXIMUM] = max(
            self._stats[STAT_CLIENTS_MAXIMUM],
            self._stats[STAT_CLIENTS_CONNECTED],
        )

    async def on_broker_client_disconnected(self, client_id: str, client_session: Session) -> None:
        """Handle broker client disconnection."""
        self._stats[STAT_CLIENTS_CONNECTED] -= 1
        self._stats[STAT_CLIENTS_DISCONNECTED] += 1

    @dataclass
    class Config:
        """Configuration struct for plugin."""

        sys_interval: int = 20



from typing_extensions import Self

from amqtt.errors import AMQTTError
from amqtt.mqtt.packet import PUBREC, MQTTFixedHeader, MQTTPacket, PacketIdVariableHeader


class PubrecPacket(MQTTPacket[PacketIdVariableHeader, None, MQTTFixedHeader]):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: PacketIdVariableHeader | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(PUBREC, 0x00)
        else:
            if fixed.packet_type is not PUBREC:
                msg = f"Invalid fixed packet type {fixed.packet_type} for PubrecPacket init"
                raise AMQTTError(msg)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id: int) -> Self:
        v_header = PacketIdVariableHeader(packet_id)
        return cls(variable_header=v_header)

    @property
    def packet_id(self) -> int:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.packet_id = val



import asyncio
from collections.abc import Generator
import contextlib
from dataclasses import dataclass
import json
import logging
import os
from pathlib import Path
import socket
import sys
from typing import Any

import typer

from amqtt import __version__ as amqtt_version
from amqtt.client import MQTTClient
from amqtt.errors import ClientError, ConnectError
from amqtt.utils import read_yaml_config

logger = logging.getLogger(__name__)


def _gen_client_id() -> str:
    pid = os.getpid()
    hostname = socket.gethostname()
    return f"amqtt_pub/{pid}-{hostname}"


def _get_extra_headers(extra_headers_json: str | None = None) -> dict[str, Any]:
    try:
        extra_headers: dict[str, Any] = json.loads(extra_headers_json or "{}")
    except (json.JSONDecodeError, TypeError):
        return {}
    return extra_headers


@dataclass
class MessageInput:
    message_str: str | None = None
    file: str | None = None
    stdin: bool | None = False
    lines: bool | None = False
    no_message: bool | None = False

    def get_message(self) -> Generator[bytes | bytearray]:
        if self.no_message:
            yield b""
        if self.message_str:
            yield self.message_str.encode(encoding="utf-8")
        if self.file:
            try:
                with Path(self.file).open(encoding="utf-8") as f:
                    for line in f:
                        yield line.encode(encoding="utf-8")
            except Exception:
                logger.exception(f"Failed to read file '{self.file}'")
        if self.lines:
            for line in sys.stdin:
                if line:
                    yield line.encode(encoding="utf-8")
        if self.stdin:
            messages = bytearray()
            for line in sys.stdin:
                messages.extend(line.encode(encoding="utf-8"))
            yield messages


@dataclass
class CAInfo:
    ca_file: str | None = None
    ca_path: str | None = None
    ca_data: str | None = None


async def do_pub(
    client: MQTTClient,
    topic: str,
    message_input: MessageInput,
    ca_info: CAInfo,
    url: str | None = None,
    clean_session: bool = False,
    retain: bool = False,
    extra_headers_json: str | None = None,
    qos: int | None = None,
) -> None:
    """Publish the message."""
    running_tasks = []

    try:
        logger.info(f"{client.client_id} Connecting to broker")

        await client.connect(
            uri=url,
            cleansession=clean_session,
            cafile=ca_info.ca_file,
            capath=ca_info.ca_path,
            cadata=ca_info.ca_data,
            additional_headers=_get_extra_headers(extra_headers_json),
        )

        for message in message_input.get_message():
            logger.info(f"{client.client_id} Publishing to '{topic}'")
            task = asyncio.ensure_future(client.publish(topic, message, qos, retain))
            running_tasks.append(task)

        if running_tasks:
            await asyncio.wait(running_tasks)

        await client.disconnect()
        logger.info(f"{client.client_id} Disconnected from broker")
    except KeyboardInterrupt:
        await client.disconnect()
        logger.info(f"{client.client_id} Disconnected from broker")
    except ConnectError as ce:
        logger.fatal(f"Connection to '{client.session.broker_uri if client.session else url}' failed")
        raise ConnectError from ce
    except asyncio.CancelledError as ce:
        logger.fatal("Publish canceled due to previous error")
        raise asyncio.CancelledError from ce

app = typer.Typer(add_completion=False, rich_markup_mode=None)


def main() -> None:
    """Entry point for the amqtt publisher."""
    app()


def _version(v: bool) -> None:
    if v:
        typer.echo(f"{amqtt_version}")
        raise typer.Exit(code=0)

@app.command()
def publisher_main(  # pylint: disable=R0914,R0917  # noqa : PLR0913
    url: str | None = typer.Option(None, "--url", help="Broker connection URL, *must conform to MQTT or URI scheme: `[mqtt(s)|ws(s)]://<username:password>@HOST:port`*"),
    config_file: str | None = typer.Option(None, "-c", "--config-file", help="Client configuration file"),
    client_id: str | None = typer.Option(None, "-i", "--client-id", help="client identification for mqtt connection. *default: process id and the hostname of the client*"),
    qos: int = typer.Option(0, "--qos", "-q", help="Quality of service (0, 1, or 2)"),
    retain: bool = typer.Option(False, "-r", help="Set retain flag on connect"),
    topic: str = typer.Option(..., "-t", "--topic", help="Message topic"),
    message: str | None = typer.Option(None, "-m", "--message", help="Message data to send"),
    file: str | None = typer.Option(None, "-f", "--file", help="Path to file, will publish each line as a separate message."),
    stdin: bool = typer.Option(False, "-s", "--stdin", help="Read from standard input, all content read is sent as a single message."),
    lines: bool = typer.Option(False, "-l", "--lines", help="Read from stdin, will publish message for each line."),
    no_message: bool = typer.Option(False, "-n", "--no-message", help="Publish an empty (null, zero length) message"),
    keep_alive: int | None = typer.Option(None, "-k", help="Keep alive timeout, in seconds."),
    clean_session: bool = typer.Option(False, "--clean-session", help="Clean session on connect. *default: False*"),
    ca_file: str | None = typer.Option(None, "--ca-file", help="Define the path to a file containing PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    ca_path: str | None = typer.Option(None, "--ca-path", help="Define the path to a directory containing PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    ca_data: str | None = typer.Option(None, "--ca-data", help="Set the PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    will_topic: str | None = typer.Option(None, "--will-topic", help="The topic on which to send a Will, in the event that the client disconnects unexpectedly."),
    will_message: str | None = typer.Option(None, "--will-message", help="Specify a message that will be stored by the broker and sent out if this client disconnects unexpectedly. *required if `--will-topic` is specified*."),
    will_qos: int = typer.Option(0, "--will-qos", help="The QoS to use for the Will. *default: 0, only valid if `--will-topic` is specified*"),
    will_retain: bool = typer.Option(False, "--will-retain", help="If the client disconnects unexpectedly the message sent out will be treated as a retained message. *only valid, if `--will-topic` is specified*"),
    extra_headers_json: str | None = typer.Option(None, "--extra-headers", help="Specify a JSON object string with key-value pairs representing additional headers that are transmitted on the initial connection. *websocket connections only*."),
    debug: bool = typer.Option(False, "-d", help="Enable debug messages"),
    version: bool = typer.Option(False, "--version", callback=_version, is_eager=True, help="Show version and exit"), # noqa : ARG001
) -> None:
    """Command-line MQTT client for publishing simple messages."""
    provided = [bool(message), bool(file), stdin, lines, no_message]
    if sum(provided) != 1:
        typer.echo("❌ You must provide exactly one of --message, --file, --stdin, --lines or --no-message", err=True)
        raise typer.Exit(code=1)

    if bool(will_message) != bool(will_topic):
        typer.echo("❌ must specify both 'will_message' and 'will_topic' ")
        raise typer.Exit(code=1)

    if will_retain and not (will_message and will_topic):
        typer.echo("❌ 'will-retain' only valid if 'will_message' and 'will_topic' are specified.", err=True)
        raise typer.Exit(code=1)

    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level, format=formatter)

    if config_file:
        config = read_yaml_config(config_file)
    else:
        default_config_path = Path(__file__).parent / "default_client.yaml"
        logger.debug(f"Using default configuration from {default_config_path}")
        config = read_yaml_config(default_config_path)

    if not client_id:
        client_id = _gen_client_id()

    if not isinstance(config, dict):
        logger.debug("Failed to correctly initialize config")
        return

    if keep_alive:
        config["keep_alive"] = int(keep_alive)

    if will_topic and will_message:
        config["will"] = {
            "topic": will_topic,
            "message": will_message,
            "qos": int(will_qos),
            "retain": will_retain,
        }

    client = MQTTClient(client_id=client_id, config=config)
    message_input = MessageInput(
        message_str=message,
        file=file,
        stdin=stdin,
        no_message=no_message,
        lines=lines,
    )
    ca_info = CAInfo(
        ca_file=ca_file,
        ca_path=ca_path,
        ca_data=ca_data,
    )
    with contextlib.suppress(KeyboardInterrupt):
        try:
            asyncio.run(
                do_pub(
                    client=client,
                    message_input=message_input,
                    url=url,
                    topic=topic,
                    retain=retain,
                    clean_session=clean_session,
                    ca_info=ca_info,
                    extra_headers_json=extra_headers_json,
                    qos=qos,
                )
            )
        except (ClientError, ConnectError) as exc:
            typer.echo("❌ Connection failed", err=True)
            raise typer.Exit(code=1) from exc


if __name__ == "__main__":
    typer.run(main)



import asyncio
from decimal import ROUND_HALF_UP, Decimal
from struct import pack, unpack

from amqtt.adapters import ReaderAdapter
from amqtt.errors import NoDataError, ZeroLengthReadError


def bytes_to_hex_str(data: bytes | bytearray) -> str:
    """Convert a sequence of bytes into its displayable hex representation, ie: 0x??????.

    :param data: byte sequence
    :return: Hexadecimal displayable representation.
    """
    return "0x" + "".join(format(b, "02x") for b in data)


def bytes_to_int(data: bytes | int) -> int:
    """Convert a sequence of bytes to an integer using big endian byte ordering.

    :param data: byte sequence
    :return: integer value.
    """
    if isinstance(data, int):
        return data

    return int.from_bytes(data, byteorder="big")


def int_to_bytes(int_value: int, length: int) -> bytes:
    """Convert an integer to a sequence of bytes using big endian byte ordering.

    :param int_value: integer value to convert
    :param length: byte length (must be 1 or 2)
    :return: byte sequence
    :raises ValueError: if the length is unsupported
    """
    # Map length to the appropriate format string
    fmt_mapping = {
        1: "!B",  # 1 byte, unsigned char
        2: "!H",  # 2 bytes, unsigned short
    }

    fmt = fmt_mapping.get(length)
    if not fmt:
        msg = "Unsupported length for int to bytes conversion. Only lengths 1 or 2 are allowed."
        raise ValueError(msg)

    return pack(fmt, int_value)


async def read_or_raise(reader: ReaderAdapter | asyncio.StreamReader, n: int = -1) -> bytes:
    """Read a given byte number from Stream. NoDataException is raised if read gives no data.

    :param reader: reader adapter
    :param n: number of bytes to read
    :return: bytes read.
    """
    try:
        data = await reader.read(n)
    except (asyncio.IncompleteReadError, ConnectionResetError, BrokenPipeError):
        data = None
    if data is None:
        msg = "No more data"
        raise NoDataError(msg)
    return data


async def decode_string(reader: ReaderAdapter | asyncio.StreamReader) -> str:
    """Read a string from a reader and decode it according to MQTT string specification.

    :param reader: Stream reader
    :return: string read from stream.
    """
    length_bytes = await read_or_raise(reader, 2)
    if len(length_bytes) < 1:
        raise ZeroLengthReadError
    str_length = unpack("!H", length_bytes)[0]
    if str_length:
        byte_str = await read_or_raise(reader, str_length)
        try:
            return byte_str.decode(encoding="utf-8")
        except UnicodeDecodeError:
            return str(byte_str)
    else:
        return ""


async def decode_data_with_length(reader: ReaderAdapter | asyncio.StreamReader) -> bytes:
    """Read data from a reader. Data is prefixed with 2 bytes length.

    :param reader: Stream reader
    :return: bytes read from stream (without length).
    """
    length_bytes = await read_or_raise(reader, 2)
    if len(length_bytes) < 1:
        raise ZeroLengthReadError
    bytes_length = unpack("!H", length_bytes)[0]
    return await read_or_raise(reader, bytes_length)


def encode_string(string: str) -> bytes:
    """Encode a string with its length as prefix.

    :param string: string to encode
    :return: string with length prefix.
    """
    data = string.encode(encoding="utf-8")
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


def encode_data_with_length(data: bytes | bytearray) -> bytes:
    """Encode data with its length as prefix.

    :param data: data to encode
    :return: data with length prefix.
    """
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


async def decode_packet_id(reader: ReaderAdapter | asyncio.StreamReader) -> int:
    """Read a packet ID as 2-bytes int from stream according to MQTT specification (2.3.1).

    :param reader: Stream reader
    :return: Packet ID.
    """
    packet_id_bytes = await read_or_raise(reader, 2)
    packet_id = unpack("!H", packet_id_bytes)
    packet: int = packet_id[0]
    return packet


def int_to_bytes_str(value: int) -> bytes:
    """Convert an int value to a bytes array containing the numeric character.

    Ex: 123 -> b'123'
    :param value: int value to convert
    :return: bytes array.
    """
    return str(value).encode("utf-8")


def float_to_bytes_str(value: float, places:int=3) -> bytes:
    """Convert an float value to a bytes array containing the numeric character."""
    quant = Decimal(f"0.{''.join(['0' for i in range(places-1)])}1")
    rounded = Decimal(value).quantize(quant, rounding=ROUND_HALF_UP)
    return str(rounded).encode("utf-8")



import asyncio
from asyncio import CancelledError, futures
from collections import deque
from collections.abc import Generator
import copy
from functools import partial
import logging
from pathlib import Path
import re
import ssl
from typing import Any, ClassVar, TypeAlias

from transitions import Machine, MachineError
import websockets.asyncio.server
from websockets.asyncio.server import ServerConnection

from amqtt.adapters import (
    ReaderAdapter,
    StreamReaderAdapter,
    StreamWriterAdapter,
    WebSocketsReader,
    WebSocketsWriter,
    WriterAdapter,
)
from amqtt.contexts import Action, BaseContext
from amqtt.errors import AMQTTError, BrokerError, MQTTError, NoDataError
from amqtt.mqtt.protocol.broker_handler import BrokerProtocolHandler
from amqtt.session import ApplicationMessage, OutgoingApplicationMessage, Session
from amqtt.utils import format_client_message, gen_client_id, read_yaml_config

from .events import BrokerEvents
from .mqtt.constants import QOS_0, QOS_1, QOS_2
from .mqtt.disconnect import DisconnectPacket
from .plugins.manager import PluginManager

_CONFIG_LISTENER: TypeAlias = dict[str, int | bool | dict[str, Any]]
_BROADCAST: TypeAlias = dict[str, Session | str | bytes | bytearray | int | None]


_defaults = read_yaml_config(Path(__file__).parent / "scripts/default_broker.yaml")


# Default port numbers
DEFAULT_PORTS = {"tcp": 1883, "ws": 8883}
AMQTT_MAGIC_VALUE_RET_SUBSCRIBED = 0x80


class RetainedApplicationMessage(ApplicationMessage):
    __slots__ = ("data", "qos", "source_session", "topic")

    def __init__(self, source_session: Session | None, topic: str, data: bytes | bytearray, qos: int | None = None) -> None:
        super().__init__(None, topic, qos, data, retain=True)
        self.source_session = source_session
        self.topic = topic
        self.data = data
        self.qos = qos


class Server:
    def __init__(
        self,
        listener_name: str,
        server_instance: asyncio.Server | websockets.asyncio.server.Server,
        max_connections: int = -1,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.instance = server_instance
        self.conn_count = 0
        self.listener_name = listener_name
        self.max_connections = max_connections
        self.semaphore = asyncio.Semaphore(max_connections) if max_connections > 0 else None

    async def acquire_connection(self) -> None:
        if self.semaphore:
            await self.semaphore.acquire()
        self.conn_count += 1
        self.logger.info(
            f"Listener '{self.listener_name}': {self.conn_count}/"
            f"{self.max_connections if self.max_connections > 0 else '∞'} connections acquired",
        )

    def release_connection(self) -> None:
        if self.semaphore:
            self.semaphore.release()
        self.conn_count -= 1
        self.logger.info(
            f"Listener '{self.listener_name}': {self.conn_count}/"
            f"{self.max_connections if self.max_connections > 0 else '∞'} connections acquired",
        )

    async def close_instance(self) -> None:
        if self.instance:
            self.instance.close()
            await self.instance.wait_closed()


class BrokerContext(BaseContext):
    """BrokerContext is used as the context passed to plugins interacting with the broker.

    It act as an adapter to broker services from plugins developed for HBMQTT broker.
    """

    def __init__(self, broker: "Broker") -> None:
        super().__init__()
        self.config: _CONFIG_LISTENER | None = None
        self._broker_instance = broker

    async def broadcast_message(self, topic: str, data: bytes, qos: int | None = None) -> None:
        await self._broker_instance.internal_message_broadcast(topic, data, qos)

    def retain_message(self, topic_name: str, data: bytes | bytearray, qos: int | None = None) -> None:
        self._broker_instance.retain_message(None, topic_name, data, qos)

    @property
    def sessions(self) -> Generator[Session]:
        for session in self._broker_instance.sessions.values():
            yield session[0]

    @property
    def retained_messages(self) -> dict[str, RetainedApplicationMessage]:
        return self._broker_instance.retained_messages

    @property
    def subscriptions(self) -> dict[str, list[tuple[Session, int]]]:
        return self._broker_instance.subscriptions


class Broker:
    """MQTT 3.1.1 compliant broker implementation.

    Args:
        config: dictionary of configuration options (see [broker configuration](broker_config.md)).
        loop: asyncio loop. defaults to `asyncio.new_event_loop()`.
        plugin_namespace: plugin namespace to use when loading plugin entry_points. defaults to `amqtt.broker.plugins`.

    Raises:
        BrokerError, ParserError, PluginError

    """

    states: ClassVar[list[str]] = [
        "new",
        "starting",
        "started",
        "not_started",
        "stopping",
        "stopped",
        "not_stopped",
    ]

    def __init__(
        self,
        config: _CONFIG_LISTENER | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
        plugin_namespace: str | None = None,
    ) -> None:
        """Initialize the broker."""
        self.logger = logging.getLogger(__name__)
        self.config = copy.deepcopy(_defaults or {})
        if config is not None:
            # if 'plugins' isn't in the config but 'auth'/'topic-check' is included, assume this is a legacy config
            if ("auth" in config or "topic-check" in config) and "plugins" not in config:
                # set to None so that the config isn't updated with the new-style default plugin list
                config["plugins"] = None  # type: ignore[assignment]
            self.config.update(config)
        self._build_listeners_config(self.config)

        self._loop = loop or asyncio.get_running_loop()
        self._servers: dict[str, Server] = {}
        self._init_states()
        self._sessions: dict[str, tuple[Session, BrokerProtocolHandler]] = {}
        self._subscriptions: dict[str, list[tuple[Session, int]]] = {}
        self._retained_messages: dict[str, RetainedApplicationMessage] = {}

        self._topic_filter_matchers: dict[str, re.Pattern[str]] = {}

        # Broadcast queue for outgoing messages
        self._broadcast_queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()
        self._broadcast_task: asyncio.Task[Any] | None = None
        self._broadcast_shutdown_waiter: asyncio.Future[Any] = futures.Future()

        # Tasks queue for managing broadcasting tasks
        self._tasks_queue: deque[asyncio.Task[OutgoingApplicationMessage]] = deque()

        # Initialize plugins manager

        context = BrokerContext(self)
        context.config = self.config
        namespace = plugin_namespace or "amqtt.broker.plugins"
        self.plugins_manager = PluginManager(namespace, context, self._loop)

    def _build_listeners_config(self, broker_config: _CONFIG_LISTENER) -> None:
        self.listeners_config = {}
        try:
            listeners_config = broker_config.get("listeners")
            if not isinstance(listeners_config, dict):
                msg = "Listener config not found or invalid"
                raise BrokerError(msg)
            defaults = listeners_config.get("default")
            if defaults is None:
                msg = "Listener config has not default included or is invalid"
                raise BrokerError(msg)

            for listener_name, listener_conf in listeners_config.items():
                config = defaults.copy()
                config.update(listener_conf)
                self.listeners_config[listener_name] = config
        except KeyError as ke:
            msg = f"Listener config not found or invalid: {ke}"
            raise BrokerError(msg) from ke

    def _init_states(self) -> None:
        self.transitions = Machine(states=Broker.states, initial="new")
        self.transitions.add_transition(trigger="start", source="new", dest="starting", before=self._log_state_change)
        self.transitions.add_transition(trigger="starting_fail", source="starting", dest="not_started")
        self.transitions.add_transition(trigger="starting_success", source="starting", dest="started")
        self.transitions.add_transition(trigger="shutdown", source="started", dest="stopping")
        self.transitions.add_transition(trigger="stopping_success", source="stopping", dest="stopped")
        self.transitions.add_transition(trigger="stopping_failure", source="stopping", dest="not_stopped")
        self.transitions.add_transition(trigger="start", source="stopped", dest="starting")

    def _log_state_change(self) -> None:
        self.logger.debug(f"State transition: {self.transitions.state}")

    async def start(self) -> None:
        """Start the broker to serve with the given configuration.

        Start method opens network sockets and will start listening for incoming connections.
        """
        try:
            self._sessions.clear()
            self._subscriptions.clear()
            self._retained_messages.clear()
            self.transitions.start()
            self.logger.debug("Broker starting")
        except (MachineError, ValueError) as exc:
            # Backwards compat: MachineError is raised by transitions < 0.5.0.
            self.logger.warning(f"[WARN-0001] Invalid method call at this moment: {exc}")
            msg = f"Broker instance can't be started: {exc}"
            raise BrokerError(msg) from exc

        await self.plugins_manager.fire_event(BrokerEvents.PRE_START)
        try:
            await self._start_listeners()
            self.transitions.starting_success()
            await self.plugins_manager.fire_event(BrokerEvents.POST_START)
            self._broadcast_task = asyncio.ensure_future(self._broadcast_loop())
            self.logger.debug("Broker started")
        except Exception as e:
            self.logger.exception("Broker startup failed")
            self.transitions.starting_fail()
            msg = f"Broker instance can't be started: {e}"
            raise BrokerError(msg) from e

    async def _start_listeners(self) -> None:
        """Start network listeners based on the configuration."""
        for listener_name, listener in self.listeners_config.items():
            if "bind" not in listener:
                self.logger.debug(f"Listener configuration '{listener_name}' is not bound")
                continue

            max_connections = listener.get("max_connections", -1)
            ssl_context = self._create_ssl_context(listener) if listener.get("ssl", False) else None

            try:
                address, port = self._split_bindaddr_port(listener["bind"], DEFAULT_PORTS[listener["type"]])
            except ValueError as e:
                msg = f"Invalid port value in bind value: {listener['bind']}"
                raise BrokerError(msg) from e

            instance = await self._create_server_instance(listener_name, listener["type"], address, port, ssl_context)
            self._servers[listener_name] = Server(listener_name, instance, max_connections)

            self.logger.info(f"Listener '{listener_name}' bind to {listener['bind']} (max_connections={max_connections})")

    def _create_ssl_context(self, listener: dict[str, Any]) -> ssl.SSLContext:
        """Create an SSL context for a listener."""
        try:
            ssl_context = ssl.create_default_context(
                ssl.Purpose.CLIENT_AUTH,
                cafile=listener.get("cafile"),
                capath=listener.get("capath"),
                cadata=listener.get("cadata"),
            )
            ssl_context.load_cert_chain(listener["certfile"], listener["keyfile"])
            ssl_context.verify_mode = ssl.CERT_OPTIONAL
        except KeyError as ke:
            msg = f"'certfile' or 'keyfile' configuration parameter missing: {ke}"
            raise BrokerError(msg) from ke
        except FileNotFoundError as fnfe:
            msg = f"Can't read cert files '{listener['certfile']}' or '{listener['keyfile']}' : {fnfe}"
            raise BrokerError(msg) from fnfe
        return ssl_context

    async def _create_server_instance(
        self,
        listener_name: str,
        listener_type: str,
        address: str | None,
        port: int,
        ssl_context: ssl.SSLContext | None,
    ) -> asyncio.Server | websockets.asyncio.server.Server:
        """Create a server instance for a listener."""
        if listener_type == "tcp":
            return await asyncio.start_server(
                partial(self.stream_connected, listener_name=listener_name),
                address,
                port,
                reuse_address=True,
                ssl=ssl_context,
            )
        if listener_type == "ws":
            return await websockets.serve(
                partial(self.ws_connected, listener_name=listener_name),
                address,
                port,
                ssl=ssl_context,
                subprotocols=[websockets.Subprotocol("mqtt")],
            )
        msg = f"Unsupported listener type: {listener_type}"
        raise BrokerError(msg)

    async def shutdown(self) -> None:
        """Stop broker instance."""
        self.logger.info("Shutting down broker...")
        # Fire broker_shutdown event to plugins
        await self.plugins_manager.fire_event(BrokerEvents.PRE_SHUTDOWN)

        # Cleanup all sessions
        for client_id in list(self._sessions.keys()):
            await self._cleanup_session(client_id)

        # Clear retained messages
        self.logger.debug(f"Clearing {len(self._retained_messages)} retained messages")
        self._retained_messages.clear()

        self.transitions.shutdown()

        await self._shutdown_broadcast_loop()

        for server in self._servers.values():
            await server.close_instance()

        if not self._broadcast_queue.empty():
            self.logger.warning(f"{self._broadcast_queue.qsize()} messages not broadcasted")
            # Clear the broadcast queue
            while not self._broadcast_queue.empty():
                self._broadcast_queue.get_nowait()

        self.logger.info("Broker closed")
        await self.plugins_manager.fire_event(BrokerEvents.POST_SHUTDOWN)
        self.transitions.stopping_success()

    async def _cleanup_session(self, client_id: str) -> None:
        """Centralized cleanup logic for a session."""
        session, handler = self._sessions.pop(client_id, (None, None))

        if handler:
            self.logger.debug(f"Stopping handler for session {client_id}")
            await self._stop_handler(handler)
        if session:
            self.logger.debug(f"Clearing all subscriptions for session {client_id}")
            await self._del_all_subscriptions(session)
            session.clear_queues()

    async def internal_message_broadcast(self, topic: str, data: bytes, qos: int | None = None) -> None:
        return await self._broadcast_message(None, topic, data, qos)

    async def ws_connected(self, websocket: ServerConnection, listener_name: str) -> None:
        await self._client_connected(listener_name, WebSocketsReader(websocket), WebSocketsWriter(websocket))

    async def stream_connected(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, listener_name: str) -> None:
        await self._client_connected(listener_name, StreamReaderAdapter(reader), StreamWriterAdapter(writer))

    async def _client_connected(self, listener_name: str, reader: ReaderAdapter, writer: WriterAdapter) -> None:
        """Handle a new client connection."""
        server = self._servers.get(listener_name)
        if not server:
            msg = f"Invalid listener name '{listener_name}'"
            raise BrokerError(msg)

        await server.acquire_connection()
        remote_info = writer.get_peer_info()
        if remote_info is None:
            self.logger.warning("Remote info could not be retrieved from peer info")
            return

        remote_address, remote_port = remote_info
        self.logger.info(f"Connection from {remote_address}:{remote_port} on listener '{listener_name}'")

        try:
            handler, client_session = await self._initialize_client_session(reader, writer, remote_address, remote_port)
        except (AMQTTError, MQTTError, NoDataError) as exc:
            self.logger.warning(f"Failed to initialize client session: {exc}")
            server.release_connection()
            return

        try:
            await self._handle_client_session(reader, writer, client_session, handler, server, listener_name)
        except (AMQTTError, MQTTError, NoDataError) as exc:
            self.logger.warning(f"Error while handling client session: {exc}")
        finally:
            self.logger.debug(f"{client_session.client_id} Client disconnected")
            server.release_connection()

    async def _initialize_client_session(
        self,
        reader: ReaderAdapter,
        writer: WriterAdapter,
        remote_address: str,
        remote_port: int,
    ) -> tuple[BrokerProtocolHandler, Session]:
        """Initialize a client session and protocol handler."""
        # Wait for first packet and expect a CONNECT
        try:
            handler, client_session = await BrokerProtocolHandler.init_from_connect(reader, writer, self.plugins_manager)
        except AMQTTError as exc:
            self.logger.warning(
                f"[MQTT-3.1.0-1] {format_client_message(address=remote_address, port=remote_port)}:"
                f" Can't read first packet as CONNECT: {exc}",
            )
            raise AMQTTError(exc) from exc
        except MQTTError as exc:
            self.logger.exception(
                f"Invalid connection from {format_client_message(address=remote_address, port=remote_port)}",
            )
            await writer.close()
            raise MQTTError(exc) from exc
        except NoDataError as exc:
            self.logger.error(  # noqa: TRY400
                f"No data from {format_client_message(address=remote_address, port=remote_port)} : {exc}",
            )
            raise AMQTTError(exc) from exc

        if client_session.clean_session:
            # Delete existing session and create a new one
            if client_session.client_id is not None and client_session.client_id != "":
                await self._delete_session(client_session.client_id)
            else:
                client_session.client_id = gen_client_id()

            client_session.parent = 0
        # Get session from cache
        elif client_session.client_id in self._sessions:
            self.logger.debug(f"Found old session {self._sessions[client_session.client_id]!r}")
            client_session, _ = self._sessions[client_session.client_id]
            client_session.parent = 1
        else:
            client_session.parent = 0

        timeout_disconnect_delay = self.config.get("timeout-disconnect-delay", 0)
        if client_session.keep_alive > 0 and isinstance(timeout_disconnect_delay, int):
            client_session.keep_alive += timeout_disconnect_delay

        self.logger.debug(f"Keep-alive timeout={client_session.keep_alive}")
        return handler, client_session

    async def _handle_client_session(
        self,
        reader: ReaderAdapter,
        writer: WriterAdapter,
        client_session: Session,
        handler: BrokerProtocolHandler,
        server: Server,
        listener_name: str,
    ) -> None:
        """Handle the lifecycle of a client session."""
        authenticated = await self._authenticate(client_session, self.listeners_config[listener_name])
        if not authenticated:
            await writer.close()
            return

        if client_session.client_id is None:
            msg = "Client ID was not correctly created/set."
            raise BrokerError(msg)

        while True:
            try:
                client_session.transitions.connect()
                break
            except (MachineError, ValueError):
                if client_session.transitions.is_connected():
                    self.logger.warning(f"Client {client_session.client_id} is already connected, performing take-over.")
                    old_session = self._sessions[client_session.client_id]
                    await old_session[1].handle_connection_closed()
                    await old_session[1].stop()
                    break
                self.logger.warning(f"Client {client_session.client_id} is reconnecting too quickly, make it wait")
                await asyncio.sleep(1)

        handler.attach(client_session, reader, writer)
        self._sessions[client_session.client_id] = (client_session, handler)

        await handler.mqtt_connack_authorize(authenticated)
        await self.plugins_manager.fire_event(BrokerEvents.CLIENT_CONNECTED,
                                              client_id=client_session.client_id,
                                              client_session=client_session)

        self.logger.debug(f"{client_session.client_id} Start messages handling")
        await handler.start()

        # publish messages that were retained because the client session was disconnected
        self.logger.debug(f"Retained messages queue size: {client_session.retained_messages.qsize()}")
        await self._publish_session_retained_messages(client_session)

        # if this is not a new session, there are subscriptions associated with them; publish any topic retained messages
        self.logger.debug("Publish retained messages to a pre-existing session's subscriptions.")
        for topic in self._subscriptions:
            await self._publish_retained_messages_for_subscription( (topic, QOS_0), client_session)



        await self._client_message_loop(client_session, handler)

    async def _client_message_loop(self, client_session: Session, handler: BrokerProtocolHandler) -> None:
        """Run the main loop to handle client messages."""
        # Init and start loop for handling client messages (publish, subscribe/unsubscribe, disconnect)
        disconnect_waiter = asyncio.ensure_future(handler.wait_disconnect())
        subscribe_waiter = asyncio.ensure_future(handler.get_next_pending_subscription())
        unsubscribe_waiter = asyncio.ensure_future(handler.get_next_pending_unsubscription())
        wait_deliver = asyncio.ensure_future(handler.mqtt_deliver_next_message())
        connected = True

        while connected:
            try:
                done, _ = await asyncio.wait(
                    [
                        disconnect_waiter,
                        subscribe_waiter,
                        unsubscribe_waiter,
                        wait_deliver,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                if disconnect_waiter in done:
                    # handle the disconnection: normal or abnormal result, either way, the client is no longer connected
                    await self._handle_disconnect(client_session, handler, disconnect_waiter)
                    connected = False

                    # no need to reschedule the `disconnect_waiter` since we're exiting the message loop


                if subscribe_waiter in done:
                    await self._handle_subscription(client_session, handler, subscribe_waiter)
                    subscribe_waiter = asyncio.ensure_future(handler.get_next_pending_subscription())
                    self.logger.debug(repr(self._subscriptions))

                if unsubscribe_waiter in done:
                    await self._handle_unsubscription(client_session, handler, unsubscribe_waiter)
                    unsubscribe_waiter = asyncio.ensure_future(handler.get_next_pending_unsubscription())

                if wait_deliver in done:
                    if not await self._handle_message_delivery(client_session, handler, wait_deliver):
                        break
                    wait_deliver = asyncio.ensure_future(handler.mqtt_deliver_next_message())

            except asyncio.CancelledError:
                self.logger.debug("Client loop cancelled")
                break

        disconnect_waiter.cancel()
        subscribe_waiter.cancel()
        unsubscribe_waiter.cancel()
        wait_deliver.cancel()

    async def _handle_disconnect(
        self,
        client_session: Session,
        handler: BrokerProtocolHandler,
        disconnect_waiter: asyncio.Future[Any],
    ) -> None:
        """Handle client disconnection.

        Args:
            client_session (Session): client session
            handler (BrokerProtocolHandler): broker protocol handler
            disconnect_waiter (asyncio.Future[Any]): future to wait for disconnection

        """
        # check the disconnected waiter result
        result = disconnect_waiter.result()
        self.logger.debug(f"{client_session.client_id} Result from wait_disconnect: {result}")
        # if the client disconnects abruptly by sending no message or the message isn't a disconnect packet
        if result is None or not isinstance(result, DisconnectPacket):
            self.logger.debug(f"Will flag: {client_session.will_flag}")
            if client_session.will_flag:
                self.logger.debug(
                    f"Client {format_client_message(client_session)} disconnected abnormally, sending will message",
                )
                await self._broadcast_message(
                    client_session,
                    client_session.will_topic,
                    client_session.will_message,
                    client_session.will_qos,
                )
                if client_session.will_retain:
                    self.retain_message(
                        client_session,
                        client_session.will_topic,
                        client_session.will_message,
                        client_session.will_qos,
                    )

        # normal or not, let's end the client's session
        self.logger.debug(f"{client_session.client_id} Disconnecting session")
        await self._stop_handler(handler)
        client_session.transitions.disconnect()
        await self.plugins_manager.fire_event(BrokerEvents.CLIENT_DISCONNECTED,
                                              client_id=client_session.client_id,
                                              client_session=client_session)


    async def _handle_subscription(
        self,
        client_session: Session,
        handler: BrokerProtocolHandler,
        subscribe_waiter: asyncio.Future[Any],
    ) -> None:
        """Handle client subscription."""
        self.logger.debug(f"{client_session.client_id} handling subscription")
        subscriptions = subscribe_waiter.result()
        return_codes = [await self._add_subscription(subscription, client_session) for subscription in subscriptions.topics]
        await handler.mqtt_acknowledge_subscription(subscriptions.packet_id, return_codes)
        for index, subscription in enumerate(subscriptions.topics):
            if return_codes[index] != AMQTT_MAGIC_VALUE_RET_SUBSCRIBED:
                await self.plugins_manager.fire_event(
                    BrokerEvents.CLIENT_SUBSCRIBED,
                    client_id=client_session.client_id,
                    topic=subscription[0],
                    qos=subscription[1],
                )
                await self._publish_retained_messages_for_subscription(subscription, client_session)

    async def _handle_unsubscription(
        self,
        client_session: Session,
        handler: BrokerProtocolHandler,
        unsubscribe_waiter: asyncio.Future[Any],
    ) -> None:
        """Handle client unsubscription."""
        self.logger.debug(f"{client_session.client_id} handling unsubscription")
        unsubscription = unsubscribe_waiter.result()
        for topic in unsubscription.topics:
            self._del_subscription(topic, client_session)
            await self.plugins_manager.fire_event(
                BrokerEvents.CLIENT_UNSUBSCRIBED,
                client_id=client_session.client_id,
                topic=topic,
            )
        await handler.mqtt_acknowledge_unsubscription(unsubscription.packet_id)

    async def _handle_message_delivery(
        self,
        client_session: Session,
        handler: BrokerProtocolHandler,
        wait_deliver: asyncio.Future[Any],
    ) -> bool:
        """Handle message delivery to the client."""
        self.logger.debug(f"{client_session.client_id} handling message delivery")
        app_message = wait_deliver.result()

        if app_message is None:
            self.logger.debug("app_message was empty!")
            return True
        if not app_message.topic:
            self.logger.warning(
                f"[MQTT-4.7.3-1] - {client_session.client_id} invalid TOPIC sent in PUBLISH message, closing connection",
            )
            return False
        if "#" in app_message.topic or "+" in app_message.topic:
            self.logger.warning(
                f"[MQTT-3.3.2-2] - {client_session.client_id} invalid TOPIC sent in PUBLISH message, closing connection",
            )
            return False
        if app_message.topic.startswith("$"):
            self.logger.warning(
                f"[MQTT-4.7.2-1] - {client_session.client_id} cannot use a topic with a leading $ character."
            )
            return False

        permitted = await self._topic_filtering(client_session, topic=app_message.topic, action=Action.PUBLISH)
        if not permitted:
            self.logger.info(f"{client_session.client_id} forbidden TOPIC {app_message.topic} sent in PUBLISH message.")
        else:
            await self.plugins_manager.fire_event(
                BrokerEvents.MESSAGE_RECEIVED,
                client_id=client_session.client_id,
                message=app_message,
            )
            await self._broadcast_message(client_session, app_message.topic, app_message.data)
            if app_message.publish_packet and app_message.publish_packet.retain_flag:
                self.retain_message(client_session, app_message.topic, app_message.data, app_message.qos)
        return True

    async def _init_handler(self, session: Session, reader: ReaderAdapter, writer: WriterAdapter) -> BrokerProtocolHandler:
        """Create a BrokerProtocolHandler and attach to a session."""
        handler = BrokerProtocolHandler(self.plugins_manager, loop=self._loop)
        handler.attach(session, reader, writer)
        return handler

    async def _stop_handler(self, handler: BrokerProtocolHandler) -> None:
        """Stop a running handler and detach if from the session."""
        try:
            await handler.stop()
        except Exception:
            self.logger.exception("Failed to stop handler")

    async def _authenticate(self, session: Session, _: dict[str, Any]) -> bool:
        """Call the authenticate method on registered plugins to test user authentication.

        User is considered authenticated if all plugins called returns True.
        Plugins authenticate() method are supposed to return :
         - True if user is authentication succeed
         - False if user authentication fails
         - None if authentication can't be achieved (then plugin result is then ignored)
        :param session:
        :return:
        """
        returns = await self.plugins_manager.map_plugin_auth(session=session)

        results = [ result for _, result in returns.items() if result is not None] if returns else []
        if len(results) < 1:
            self.logger.debug("Authentication failed: no plugin responded with a boolean")
            return False

        if all(results):
            self.logger.debug("Authentication succeeded")
            return True

        for plugin, result in returns.items():
            self.logger.debug(f"Authentication '{plugin.__class__.__name__}' result: {result}")

        return False

    def retain_message(
        self,
        source_session: Session | None,
        topic_name: str | None,
        data: bytes | bytearray | None,
        qos: int | None = None,
    ) -> None:
        if data and topic_name is not None:
            # If retained flag set, store the message for further subscriptions
            self.logger.debug(f"Retaining message on topic {topic_name}")
            self._retained_messages[topic_name] = RetainedApplicationMessage(source_session, topic_name, data, qos)
        # [MQTT-3.3.1-10]
        elif topic_name in self._retained_messages:
            self.logger.debug(f"Clearing retained messages for topic '{topic_name}'")
            del self._retained_messages[topic_name]

    async def _add_subscription(self, subscription: tuple[str, int], session: Session) -> int:
        topic_filter, qos = subscription
        if "#" in topic_filter and not topic_filter.endswith("#"):
            # [MQTT-4.7.1-2] Wildcard character '#' is only allowed as last character in filter
            return 0x80
        if topic_filter != "+" and "+" in topic_filter and ("/+" not in topic_filter and "+/" not in topic_filter):
            # [MQTT-4.7.1-3] + wildcard character must occupy entire level
            return 0x80
        # Check if the client is authorised to connect to the topic
        if not await self._topic_filtering(session, topic_filter, Action.SUBSCRIBE):
            return 0x80

        # Ensure "max-qos" is an integer before using it
        max_qos = self.config.get("max-qos", qos)
        if not isinstance(max_qos, int):
            max_qos = qos

        qos = min(qos, max_qos)
        if topic_filter not in self._subscriptions:
            self._subscriptions[topic_filter] = []
        if all(s.client_id != session.client_id for s, _ in self._subscriptions[topic_filter]):
            self._subscriptions[topic_filter].append((session, qos))
        else:
            self.logger.debug(f"Client {format_client_message(session=session)} has already subscribed to {topic_filter}")
        return qos

    async def _topic_filtering(self, session: Session, topic: str, action: Action) -> bool:
        """Call the topic_filtering method on registered plugins to check that the subscription is allowed.

        User is considered allowed if all plugins called return True.
        Plugins topic_filtering() method are supposed to return :
         - True if MQTT client can be subscribed to the topic
         - False if MQTT client is not allowed to subscribe to the topic
         - None if topic filtering can't be achieved (then plugin result is then ignored)
        :param session:
        :param topic: Topic in which the client wants to subscribe / publish
        :param action: What is being done with the topic?  subscribe or publish
        :return:
        """
        if not self.plugins_manager.is_topic_filtering_enabled():
            return True

        results = await self.plugins_manager.map_plugin_topic(session=session, topic=topic, action=action)
        return all(result for result in results.values())

    async def _delete_session(self, client_id: str) -> None:
        """Delete an existing session data, for example due to clean session set in CONNECT."""
        session = self._sessions.pop(client_id, (None, None))[0]

        if session is None:
            self.logger.debug(f"Delete session : session {client_id} doesn't exist")
            return
        self.logger.debug(f"Deleted existing session {session!r}")

        # Delete subscriptions
        self.logger.debug(f"Deleting session {session!r} subscriptions")
        await self._del_all_subscriptions(session)
        session.clear_queues()

    async def _del_all_subscriptions(self, session: Session) -> None:
        """Delete all topic subscriptions for a given session."""
        filter_queue: deque[str] = deque()
        for topic in self._subscriptions:
            if self._del_subscription(topic, session):
                filter_queue.append(topic)
        for topic in filter_queue:
            if not self._subscriptions[topic]:
                del self._subscriptions[topic]

    def _del_subscription(self, a_filter: str, session: Session) -> int:
        """Delete a session subscription on a given topic.

        :param a_filter: The topic filter for the subscription.
        :param session: The session to be unsubscribed.
        :return: The number of deleted subscriptions (0 or 1).
        """
        deleted = 0
        try:
            subscriptions = self._subscriptions[a_filter]
            for index, (sub_session, _qos) in enumerate(subscriptions):
                if sub_session.client_id == session.client_id:
                    self.logger.debug(
                        f"Removing subscription on topic '{a_filter}' for client {format_client_message(session=session)}",
                    )
                    subscriptions.pop(index)
                    deleted += 1
                    break
        except KeyError:
            self.logger.debug(f"Unsubscription on topic '{a_filter}' for client {format_client_message(session=session)}")

        return deleted

    async def _broadcast_loop(self) -> None:
        """Run the main loop to broadcast messages."""
        running_tasks: deque[asyncio.Task[OutgoingApplicationMessage]] = self._tasks_queue

        try:
            while True:
                while running_tasks and running_tasks[0].done():
                    task = running_tasks.popleft()
                    try:
                        task.result()
                    except CancelledError:
                        self.logger.info(f"Task has been cancelled: {task}")
                    except Exception:
                        self.logger.exception(f"Task failed and will be skipped: {task}")

                run_broadcast_task = asyncio.ensure_future(self._run_broadcast(running_tasks))

                completed, _ = await asyncio.wait(
                    [run_broadcast_task, self._broadcast_shutdown_waiter],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                # Shutdown has been triggered by the broker, so stop the loop execution
                if self._broadcast_shutdown_waiter in completed:
                    run_broadcast_task.cancel()
                    break

        except BaseException:
            self.logger.exception("Broadcast loop stopped by exception")
            raise
        finally:
            # Wait until current broadcasting tasks end
            if running_tasks:
                await asyncio.gather(*running_tasks)

    async def _run_broadcast(self, running_tasks: deque[asyncio.Task[OutgoingApplicationMessage]]) -> None:
        """Process a single broadcast message."""
        broadcast = await self._broadcast_queue.get()

        self.logger.debug(f"Processing broadcast message: {broadcast}")

        for k_filter, subscriptions in self._subscriptions.items():

            # Skip all subscriptions which do not match the topic
            if not self._matches(broadcast["topic"], k_filter):
                self.logger.debug(f"Topic '{broadcast['topic']}' does not match filter '{k_filter}'")
                continue

            for target_session, sub_qos in subscriptions:
                qos = broadcast.get("qos", sub_qos)

                # Retain all messages which cannot be broadcasted, due to the session not being connected
                #  but only when clean session is false and qos is 1 or 2 [MQTT 3.1.2.4]
                #  and, if a client used anonymous authentication, there is no expectation that messages should be retained
                if (target_session.transitions.state != "connected"
                        and not target_session.clean_session
                        and qos in (QOS_1, QOS_2)
                        and not target_session.is_anonymous):
                    self.logger.debug(f"Session {target_session.client_id} is not connected, retaining message.")
                    await self._retain_broadcast_message(broadcast, qos, target_session)
                    continue

                # Only broadcast the message to connected clients
                if target_session.transitions.state != "connected":
                    continue

                self.logger.debug(
                    f"Broadcasting message from {format_client_message(session=broadcast['session'])}"
                    f" on topic '{broadcast['topic']}' to {format_client_message(session=target_session)}",
                )

                handler = self._get_handler(target_session)
                if handler:
                    task = asyncio.ensure_future(
                        handler.mqtt_publish(
                            broadcast["topic"],
                            broadcast["data"],
                            qos,
                            retain=False,
                        ),
                    )
                    running_tasks.append(task)

    async def _retain_broadcast_message(self, broadcast: dict[str, Any], qos: int, target_session: Session) -> None:
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                f"retaining application message from {format_client_message(session=broadcast['session'])}"
                f" on topic '{broadcast['topic']}' to client '{format_client_message(session=target_session)}'",
            )

        retained_message = RetainedApplicationMessage(broadcast["session"], broadcast["topic"], broadcast["data"], qos)
        await target_session.retained_messages.put(retained_message)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"target_session.retained_messages={target_session.retained_messages.qsize()}")

    async def _shutdown_broadcast_loop(self) -> None:
        if self._broadcast_task and not self._broadcast_shutdown_waiter.done():
            self._broadcast_shutdown_waiter.set_result(True)
            try:
                await asyncio.wait_for(self._broadcast_task, timeout=30)
            except TimeoutError as e:
                self.logger.warning(f"Failed to cleanly shutdown broadcast loop: {e}")

        if not self._broadcast_queue.empty():
            self.logger.warning(f"{self._broadcast_queue.qsize()} messages not broadcasted")

        self._broadcast_shutdown_waiter = asyncio.Future()

    async def _broadcast_message(
        self,
        session: Session | None,
        topic: str | None,
        data: bytes | bytearray | None,
        force_qos: int | None = None,
    ) -> None:
        broadcast: _BROADCAST = {"session": session, "topic": topic, "data": data}
        if force_qos is not None:
            broadcast["qos"] = force_qos
        await self._broadcast_queue.put(broadcast)

    async def _publish_session_retained_messages(self, session: Session) -> None:
        self.logger.debug(
            f"Publishing {session.retained_messages.qsize()}"
            f" messages retained for session {format_client_message(session=session)}",
        )
        publish_tasks = []
        handler = self._get_handler(session)
        if handler:
            while not session.retained_messages.empty():
                retained = await session.retained_messages.get()
                publish_tasks.append(
                    asyncio.ensure_future(
                        handler.mqtt_publish(retained.topic, retained.data, retained.qos, retain=True),
                    ),
                )
        if publish_tasks:
            await asyncio.wait(publish_tasks)

    async def _publish_retained_messages_for_subscription(self, subscription: tuple[str, int], session: Session) -> None:
        self.logger.debug(
            f"Begin broadcasting messages retained due to subscription on '{subscription[0]}'"
            f" from {format_client_message(session=session)}",
        )
        publish_tasks = []

        topic_filter, qos = subscription
        for topic, retained in self._retained_messages.items():
            self.logger.debug(f"matching : {topic} {topic_filter}")
            if self._matches(topic, topic_filter):
                self.logger.debug(f"{topic} and {topic_filter} match")
                handler = self._get_handler(session)
                if handler:
                    publish_tasks.append(
                        asyncio.Task(
                            handler.mqtt_publish(retained.topic, retained.data, min(qos, retained.qos or qos), retain=True),
                        ),
                    )
        if publish_tasks:
            await asyncio.wait(publish_tasks)
        self.logger.debug(
            f"End broadcasting messages retained due to subscription on '{subscription[0]}'"
            f" from {format_client_message(session=session)}",
        )

    def _matches(self, topic: str, a_filter: str) -> bool:
        if topic.startswith("$") and (a_filter.startswith(("+", "#"))):
            self.logger.debug("[MQTT-4.7.2-1] - ignoring broadcasting $ topic to subscriptions starting with + or #")
            return False

        if "#" not in a_filter and "+" not in a_filter:
            # if filter doesn't contain wildcard, return exact match
            return a_filter == topic

        # else use regex (re.compile is an expensive operation, store the matcher for future use)
        if a_filter not in self._topic_filter_matchers:
            self._topic_filter_matchers[a_filter] = re.compile(re.escape(a_filter)
                                                               .replace("\\#", "?.*")
                                                               .replace("\\+", "[^/]*")
                                                               .lstrip("?"))
        match_pattern = self._topic_filter_matchers[a_filter]
        return bool(match_pattern.fullmatch(topic))

    def _get_handler(self, session: Session) -> BrokerProtocolHandler | None:
        client_id = session.client_id
        if client_id:
            return self._sessions.get(client_id, (None, None))[1]
        return None

    @classmethod
    def _split_bindaddr_port(cls, port_str: str, default_port: int) -> tuple[str | None, int]:
        """Split an address:port pair into separate IP address and port. with IPv6 special-case handling.

        - Address can be specified using one of the following methods:
        - empty string      - all interfaces default port
        - 1883              - Port number only (listen all interfaces)
        - :1883             - Port number only (listen all interfaces)
        - 0.0.0.0:1883      - IPv4 address
        - [::]:1883         - IPv6 address
        """

        def _parse_port(port_str: str) -> int:
            port_str = port_str.removeprefix(":")

            if not port_str:
                return default_port

            return int(port_str)

        if port_str.startswith("["):  # IPv6 literal
            try:
                addr_end = port_str.index("]")
            except ValueError as e:
                msg = "Expecting '[' to be followed by ']'"
                raise ValueError(msg) from e

            return (port_str[0 : addr_end + 1], _parse_port(port_str[addr_end + 1 :]))

        if ":" in port_str:
            address, port_str = port_str.rsplit(":", 1)
            return (address or None, _parse_port(port_str))

        try:
            return (None, _parse_port(port_str))
        except ValueError:
            return (port_str, default_port)

    @property
    def subscriptions(self) -> dict[str, list[tuple[Session, int]]]:
        return self._subscriptions

    @property
    def retained_messages(self) -> dict[str, RetainedApplicationMessage]:
        return self._retained_messages

    @property
    def sessions(self) -> dict[str, tuple[Session, BrokerProtocolHandler]]:
        return self._sessions



from asyncio import Queue
from collections import OrderedDict
from typing import Any, ClassVar

from transitions import Machine

from amqtt.errors import AMQTTError
from amqtt.mqtt.publish import PublishPacket

OUTGOING = 0
INCOMING = 1


class ApplicationMessage:
    """ApplicationMessage and subclasses are used to store published message information flow.

    These objects can contain different information depending on the way they were created (incoming or outgoing)
    and the quality of service used between peers.
    """

    __slots__ = (
        "data",
        "packet_id",
        "puback_packet",
        "pubcomp_packet",
        "publish_packet",
        "pubrec_packet",
        "pubrel_packet",
        "qos",
        "retain",
        "topic",
    )

    def __init__(self, packet_id: int | None, topic: str, qos: int | None, data: bytes | bytearray, retain: bool) -> None:
        self.packet_id: int | None = packet_id
        """ Publish message packet identifier
            <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718025>_
        """

        self.topic: str = topic
        """ Publish message topic"""

        self.qos: int | None = qos
        """ Publish message Quality of Service"""

        self.data: bytes | bytearray = data
        """ Publish message payload data"""

        self.retain: bool = retain
        """ Publish message retain flag"""

        self.publish_packet: PublishPacket | None = None
        """ :class:`amqtt.mqtt.publish.PublishPacket` instance corresponding to the
        `PUBLISH <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037>`_ packet in the messages flow.
        ``None`` if the PUBLISH packet has not already been received or sent."""

        self.puback_packet: Any | None = None
        """ :class:`amqtt.mqtt.puback.PubackPacket` instance corresponding to the
        `PUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718043>`_ packet in the messages flow.
        ``None`` if QoS != QOS_1 or if the PUBACK packet has not already been received or sent."""

        self.pubrec_packet: Any | None = None
        """ :class:`amqtt.mqtt.puback.PubrecPacket` instance corresponding to the
        `PUBREC <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718048>`_ packet in the messages flow.
        ``None`` if QoS != QOS_2 or if the PUBREC packet has not already been received or sent."""

        self.pubrel_packet: Any | None = None
        """ :class:`amqtt.mqtt.puback.PubrelPacket` instance corresponding to the
        `PUBREL <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718053>`_ packet in the messages flow.
        ``None`` if QoS != QOS_2 or if the PUBREL packet has not already been received or sent."""

        self.pubcomp_packet: Any | None = None
        """ :class:`amqtt.mqtt.puback.PubrelPacket` instance corresponding to the
        `PUBCOMP <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718058>`_ packet in the messages flow.
        ``None`` if QoS != QOS_2 or if the PUBCOMP packet has not already been received or sent."""

    def build_publish_packet(self, dup: bool = False) -> PublishPacket:
        """Build :class:`amqtt.mqtt.publish.PublishPacket` from attributes.

        :param dup: force dup flag
        :return: :class:`amqtt.mqtt.publish.PublishPacket` built from ApplicationMessage instance attributes
        """
        return PublishPacket.build(self.topic, bytes(self.data), self.packet_id, dup, self.qos, self.retain)

    def __eq__(self, other: object) -> bool:
        """Compare two ApplicationMessage instances based on their packet_id.

        This method is used to check if two messages are the same based on their packet_id.
        :param other: The other ApplicationMessage instance to compare with.
        :return: True if the packet_id of both messages are equal, False otherwise.
        """
        if not isinstance(other, ApplicationMessage):
            return False
        return self.packet_id == other.packet_id


class IncomingApplicationMessage(ApplicationMessage):
    """Incoming :class:~amqtt.session.ApplicationMessage."""

    __slots__ = ("direction",)

    def __init__(self, packet_id: int | None, topic: str, qos: int | None, data: bytes, retain: bool) -> None:
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction: int = INCOMING


class OutgoingApplicationMessage(ApplicationMessage):
    """Outgoing :class:~amqtt.session.ApplicationMessage."""

    __slots__ = ("direction",)

    def __init__(self, packet_id: int | None, topic: str, qos: int | None, data: bytes | bytearray, retain: bool) -> None:
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction: int = OUTGOING


class Session:
    states: ClassVar[list[str]] = ["new", "connected", "disconnected"]

    def __init__(self) -> None:
        self._init_states()
        self.remote_address: str | None = None
        self.remote_port: int | None = None
        self.client_id: str | None = None
        self.clean_session: bool | None = None
        self.will_flag: bool = False
        self.will_message: bytes | bytearray | None = None
        self.will_qos: int | None = None
        self.will_retain: bool | None = None
        self.will_topic: str | None = None
        self.keep_alive: int = 0
        self.publish_retry_delay: int = 0
        self.broker_uri: str | None = None
        self.username: str | None = None
        self.password: str | None = None
        self.cafile: str | None = None
        self.capath: str | None = None
        self.cadata: bytes | None = None
        self._packet_id: int = 0
        self.parent: int = 0

        # Used to store outgoing ApplicationMessage while publish protocol flows
        self.inflight_out: OrderedDict[int, OutgoingApplicationMessage] = OrderedDict()

        # Used to store incoming ApplicationMessage while publish protocol flows
        self.inflight_in: OrderedDict[int, IncomingApplicationMessage] = OrderedDict()

        # Stores messages retained for this session (specifically when the client is disconnected)
        self.retained_messages: Queue[ApplicationMessage] = Queue()

        # Stores PUBLISH messages ID received in order and ready for application process
        self.delivered_message_queue: Queue[ApplicationMessage] = Queue()

        # identify anonymous client sessions or clients which didn't identify themselves
        self.is_anonymous: bool = False

    def _init_states(self) -> None:
        self.transitions = Machine(states=Session.states, initial="new")
        self.transitions.add_transition(
            trigger="connect",
            source="new",
            dest="connected",
        )
        self.transitions.add_transition(
            trigger="connect",
            source="disconnected",
            dest="connected",
        )
        self.transitions.add_transition(
            trigger="disconnect",
            source="connected",
            dest="disconnected",
        )
        self.transitions.add_transition(
            trigger="disconnect",
            source="new",
            dest="disconnected",
        )
        self.transitions.add_transition(
            trigger="disconnect",
            source="disconnected",
            dest="disconnected",
        )

    @property
    def next_packet_id(self) -> int:
        self._packet_id = (self._packet_id % 65535) + 1
        limit = self._packet_id
        while self._packet_id in self.inflight_in or self._packet_id in self.inflight_out:
            self._packet_id = (self._packet_id % 65535) + 1
            if self._packet_id == limit:
                msg = "More than 65535 messages pending. No free packet ID"
                raise AMQTTError(msg)

        return self._packet_id

    @property
    def inflight_in_count(self) -> int:
        return len(self.inflight_in)

    @property
    def inflight_out_count(self) -> int:
        return len(self.inflight_out)

    @property
    def retained_messages_count(self) -> int:
        return self.retained_messages.qsize()

    def __repr__(self) -> str:
        """Return a string representation of the session.

        This method is used for debugging and logging purposes.
        It includes the client ID and the current state of the session.
        """
        return type(self).__name__ + f"(clientId={self.client_id}, state={self.transitions.state})"

    def __getstate__(self) -> dict[str, Any]:
        """Return the state of the session for pickling.

        This method is called when pickling the session object.
        It returns a dictionary containing the session's state, excluding
        unpicklable entries.
        """
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state["retained_messages"]
        del state["delivered_message_queue"]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Restore the session from its state.

        This method is called when unpickling the session object.
        It restores the session's state and reinitializes the queues.
        """
        self.__dict__.update(state)
        self.retained_messages = Queue()
        self.delivered_message_queue = Queue()

    def clear_queues(self) -> None:
        """Clear all message queues associated with the session."""
        while not self.retained_messages.empty():
            self.retained_messages.get_nowait()
        while not self.delivered_message_queue.empty():
            self.delivered_message_queue.get_nowait()

    def __eq__(self, other: object) -> bool:
        """Compare two Session instances based on their client_id."""
        if not isinstance(other, Session):
            return False
        return self.client_id == other.client_id



import json
import sqlite3
from typing import Any

from amqtt.contexts import BaseContext
from amqtt.session import Session


class SQLitePlugin:
    def __init__(self, context: BaseContext) -> None:
        self.context: BaseContext = context
        self.conn: sqlite3.Connection | None = None
        self.cursor: sqlite3.Cursor | None = None
        self.db_file: str | None = None
        self.persistence_config: dict[str, Any]

        if (
            persistence_config := self.context.config.get("persistence") if self.context.config is not None else None
        ) is not None:
            self.persistence_config = persistence_config
            self.init_db()
        else:
            self.context.logger.warning("'persistence' section not found in context configuration")

    def init_db(self) -> None:
        self.db_file = self.persistence_config.get("file")
        if not self.db_file:
            self.context.logger.warning("'file' persistence parameter not found")
        else:
            try:
                self.conn = sqlite3.connect(self.db_file)
                self.cursor = self.conn.cursor()
                self.context.logger.info(f"Database file '{self.db_file}' opened")
            except Exception:
                self.context.logger.exception(f"Error while initializing database '{self.db_file}'")
        if self.cursor:
            self.cursor.execute(
                "CREATE TABLE IF NOT EXISTS session(client_id TEXT PRIMARY KEY, data BLOB)",
            )
            self.cursor.execute("PRAGMA table_info(session)")
            columns = {col[1] for col in self.cursor.fetchall()}
            required_columns = {"client_id", "data"}
            if not required_columns.issubset(columns):
                self.context.logger.error("Database schema for 'session' table is incompatible.")

    async def save_session(self, session: Session) -> None:
        if self.cursor and self.conn:
            dump: str = json.dumps(session, default=str)
            try:
                self.cursor.execute(
                    "INSERT OR REPLACE INTO session (client_id, data) VALUES (?, ?)",
                    (session.client_id, dump),
                )
                self.conn.commit()
            except Exception:
                self.context.logger.exception(f"Failed saving session '{session}'")

    async def find_session(self, client_id: str) -> Session | None:
        if self.cursor:
            row = self.cursor.execute(
                "SELECT data FROM session where client_id=?",
                (client_id,),
            ).fetchone()
            return json.loads(row[0]) if row else None
        return None

    async def del_session(self, client_id: str) -> None:
        if self.cursor and self.conn:
            try:
                exists = self.cursor.execute("SELECT 1 FROM session WHERE client_id=?", (client_id,)).fetchone()
                if exists:
                    self.cursor.execute("DELETE FROM session where client_id=?", (client_id,))
                    self.conn.commit()
            except Exception:
                self.context.logger.exception(f"Failed deleting session with client_id '{client_id}'")

    async def on_broker_post_shutdown(self) -> None:
        if self.conn:
            try:
                self.conn.close()
                self.context.logger.info(f"Database file '{self.db_file}' closed")
            except Exception:
                self.context.logger.exception("Error closing database connection")
            finally:
                self.conn = None



from enum import Enum
import logging
from typing import TYPE_CHECKING, Any

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import asyncio


class BaseContext:
    def __init__(self) -> None:
        self.loop: asyncio.AbstractEventLoop | None = None
        self.logger: logging.Logger = _LOGGER
        self.config: dict[str, Any] | None = None


class Action(Enum):
    """Actions issued by the broker."""

    SUBSCRIBE = "subscribe"
    PUBLISH = "publish"



from dataclasses import dataclass, field
from typing import Any

from amqtt.contexts import Action, BaseContext
from amqtt.plugins.base import BaseTopicPlugin
from amqtt.session import Session


class TopicTabooPlugin(BaseTopicPlugin):
    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)
        self._taboo: list[str] = ["prohibited", "top-secret", "data/classified"]

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool:
        filter_result = await super().topic_filtering(session=session, topic=topic, action=action)
        if filter_result:
            if session and session.username == "admin":
                return True
            return not (topic and topic in self._taboo)
        return filter_result


class TopicAccessControlListPlugin(BaseTopicPlugin):

    @staticmethod
    def topic_ac(topic_requested: str, topic_allowed: str) -> bool:
        req_split = topic_requested.split("/")
        allowed_split = topic_allowed.split("/")
        ret = True
        for i in range(max(len(req_split), len(allowed_split))):
            try:
                a_aux = req_split[i]
                b_aux = allowed_split[i]
            except IndexError:
                ret = False
                break
            if b_aux == "#":
                break
            if b_aux in ("+", a_aux):
                continue
            ret = False
            break
        return ret

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool:
        filter_result = await super().topic_filtering(session=session, topic=topic, action=action)
        if not filter_result:
            return False

        # hbmqtt and older amqtt do not support publish filtering
        if action == Action.PUBLISH and not self._get_config_option("publish-acl", {}):
            # maintain backward compatibility, assume permitted
            return True

        req_topic = topic
        if not req_topic:
            return False\

        username = session.username if session else None
        if username is None:
            username = "anonymous"

        acl: dict[str, Any] = {}
        match action:
            case Action.PUBLISH:
                acl = self._get_config_option("publish-acl", {})
            case Action.SUBSCRIBE:
                acl = self._get_config_option("acl", {})

        allowed_topics = acl.get(username, [])
        if not allowed_topics:
            return False

        return any(self.topic_ac(req_topic, allowed_topic) for allowed_topic in allowed_topics)

    @dataclass
    class Config:
        """Mappings of username and list of approved topics."""

        publish_acl: dict[str, list[str]] = field(default_factory=dict)
        acl: dict[str, list[str]] = field(default_factory=dict)



import asyncio
from collections import deque
from collections.abc import Callable, Coroutine
import contextlib
import copy
from functools import wraps
import logging
from pathlib import Path
import ssl
from typing import TYPE_CHECKING, Any, TypeAlias, cast
from urllib.parse import urlparse, urlunparse

import websockets
from websockets import HeadersLike, InvalidHandshake, InvalidURI

from amqtt.adapters import (
    StreamReaderAdapter,
    StreamWriterAdapter,
    WebSocketsReader,
    WebSocketsWriter,
)
from amqtt.contexts import BaseContext
from amqtt.errors import ClientError, ConnectError, ProtocolHandlerError
from amqtt.mqtt.connack import CONNECTION_ACCEPTED
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.protocol.client_handler import ClientProtocolHandler
from amqtt.plugins.manager import PluginManager
from amqtt.session import ApplicationMessage, OutgoingApplicationMessage, Session
from amqtt.utils import gen_client_id, read_yaml_config

if TYPE_CHECKING:
    from websockets.asyncio.client import ClientConnection

_defaults: dict[str, Any] | None = read_yaml_config(Path(__file__).parent / "scripts/default_client.yaml")


class ClientContext(BaseContext):
    """ClientContext is used as the context passed to plugins interacting with the client.

    It acts as an adapter to client services from plugins.
    """

    def __init__(self) -> None:
        super().__init__()
        self.config = None


base_logger = logging.getLogger(__name__)

_F: TypeAlias = Callable[..., Coroutine[Any, Any, Any]]


def mqtt_connected(func: _F) -> _F:
    """MQTTClient coroutines decorator which will wait until connection before calling the decorated method.

    :param func: coroutine to be called once connected
    :return: coroutine result.
    """

    @wraps(func)
    async def wrapper(self: "MQTTClient", *args: Any, **kwargs: Any) -> Any:
        if not self._connected_state.is_set():
            base_logger.warning("Client not connected, waiting for it")
            _, pending = await asyncio.wait(
                [
                    asyncio.create_task(self._connected_state.wait()),
                    asyncio.create_task(self._no_more_connections.wait()),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for t in pending:
                t.cancel()
            if self._no_more_connections.is_set():
                msg = "Will not reconnect"
                raise ClientError(msg)
        return await func(self, *args, **kwargs)

    return cast("_F", wrapper)


class MQTTClient:
    """MQTT client implementation.

    MQTTClient instances provides API for connecting to a broker and send/receive
     messages using the MQTT protocol.

    Args:
        client_id: MQTT client ID to use when connecting to the broker. If none,
            it will be generated randomly by `amqtt.utils.gen_client_id`
        config: dictionary of configuration options (see [client configuration](client_config.md)).

    Raises:
        PluginError

    """

    def __init__(self, client_id: str | None = None, config: dict[str, Any] | None = None) -> None:
        self.logger = logging.getLogger(__name__)
        self.config = copy.deepcopy(_defaults or {})
        if config is not None:
            self.config.update(config)
        self.client_id = client_id if client_id is not None else gen_client_id()

        self.session: Session | None = None
        self._handler: ClientProtocolHandler | None = None
        self._disconnect_task: asyncio.Task[Any] | None = None
        self._connected_state = asyncio.Event()
        self._no_more_connections = asyncio.Event()
        self.additional_headers: dict[str, Any] | HeadersLike = {}

        # Init plugins manager
        context = ClientContext()
        context.config = self.config
        self.plugins_manager: PluginManager[ClientContext] = PluginManager("amqtt.client.plugins", context)
        self.client_tasks: deque[asyncio.Task[Any]] = deque()

    async def connect(
        self,
        uri: str | None = None,
        cleansession: bool | None = None,
        cafile: str | None = None,
        capath: str | None = None,
        cadata: str | None = None,
        additional_headers: dict[str, Any] | HeadersLike | None = None,
    ) -> int:
        """Connect to a remote broker.

        At first, a network connection is established with the server
        using the given protocol (``mqtt``, ``mqtts``, ``ws`` or ``wss``).
        Once the socket is connected, a
        [CONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>)
        message is sent with the requested information.

        Args:
            uri: Broker URI connection, conforming to
                [MQTT URI scheme](https://github.com/mqtt/mqtt.github.io/wiki/URI-Scheme). default,
                will be taken from the ``uri`` config attribute.
            cleansession: MQTT CONNECT clean session flag
            cafile: server certificate authority file (optional, used for secured connection)
            capath: server certificate authority path (optional, used for secured connection)
            cadata: server certificate authority data (optional, used for secured connection)
            additional_headers: a dictionary with additional http headers that should be sent on the
                initial connection (optional, used only with websocket connections)

        Returns:
            [CONNACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033)'s return code

        Raises:
            ClientError, ConnectError

        """
        additional_headers = additional_headers if additional_headers is not None else {}
        self.session = self._init_session(uri, cleansession, cafile, capath, cadata)
        self.additional_headers = additional_headers
        self.logger.debug(f"Connecting to: {self.session.broker_uri}")

        try:
            return await self._do_connect()
        except asyncio.CancelledError as e:
            msg = "Future or Task was cancelled"
            raise ConnectError(msg) from e
        except Exception as e:
            self.logger.warning(f"Connection failed: {e!r}")
            if not self.config.get("auto_reconnect", False):
                raise
            return await self.reconnect()

    async def disconnect(self) -> None:
        """Disconnect from the connected broker.

        This method sends a [DISCONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718090)
        message and closes the network socket.

        """
        await self.cancel_tasks()

        if not (self.session and self._handler):
            self.logger.warning("Session or handler not initialized, ignoring disconnect.")
            return

        if not self.session.transitions.is_connected():
            self.logger.warning("Client session not connected, ignoring call.")
            return

        if self._disconnect_task and not self._disconnect_task.done():
            self._disconnect_task.cancel()

        await self._handler.mqtt_disconnect()
        self._connected_state.clear()
        await self._handler.stop()
        self.session.transitions.disconnect()

    async def cancel_tasks(self) -> None:
        """Cancel all pending tasks."""
        while self.client_tasks:
            task = self.client_tasks.pop()
            task.cancel()

    async def reconnect(self, cleansession: bool | None = None) -> int:
        """Reconnect a previously connected broker.

        Reconnection tries to establish a network connection
        and send a [CONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028) message.
        Retries interval and attempts can be controlled with the ``reconnect_max_interval``
        and ``reconnect_retries`` configuration parameters.

        Args:
            cleansession: clean session flag used in MQTT CONNECT messages sent for reconnections.

        Returns:
            [CONNACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033) return code

        Raises:
            amqtt.client.ConnectException: if re-connection fails after max retries.

        """
        if self.session and self.session.transitions.is_connected():
            self.logger.warning("Client already connected")
            return CONNECTION_ACCEPTED

        if self.session and cleansession:
            self.session.clean_session = cleansession
        self.logger.debug(f"Reconnecting with session parameters: {self.session}")

        reconnect_max_interval = self.config.get("reconnect_max_interval", 10)
        reconnect_retries = self.config.get("reconnect_retries", 2)
        nb_attempt = 1

        while True:
            try:
                self.logger.debug(f"Reconnect attempt {nb_attempt}...")
                return await self._do_connect()
            except asyncio.CancelledError as e:
                msg = "Future or Task was cancelled"
                raise ConnectError(msg) from e
            except Exception as e:
                self.logger.warning(f"Reconnection attempt failed: {e!r}")
                self.logger.debug("", exc_info=True)
                if 0 <= reconnect_retries < nb_attempt:
                    self.logger.exception("Maximum connection attempts reached. Reconnection aborted.")
                    self.logger.debug("", exc_info=True)
                    msg = "Too many failed attempts"
                    raise ConnectError(msg) from e
                delay = min(reconnect_max_interval, 2**nb_attempt)
                self.logger.debug(f"Waiting {delay} seconds before next attempt")
                await asyncio.sleep(delay)
                nb_attempt += 1

    async def _do_connect(self) -> int:
        return_code = await self._connect_coro()
        self._disconnect_task = asyncio.create_task(self.handle_connection_close())
        return return_code

    @mqtt_connected
    async def ping(self) -> None:
        """Ping the broker.

        Send a MQTT [PINGREQ](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718081)
        message for response.

        """
        if self.session and self._handler and self.session.transitions.is_connected():
            await self._handler.mqtt_ping()
        elif not self.session:
            self.logger.warning("Session is not initialized.")
        elif not self._handler:
            self.logger.warning("Handler is not initialized.")
        else:
            self.logger.warning(f"PING incompatible with state '{self.session.transitions.state}'")

    @mqtt_connected
    async def publish(
        self,
        topic: str,
        message: bytes,
        qos: int | None = None,
        retain: bool | None = None,
        ack_timeout: int | None = None,
    ) -> OutgoingApplicationMessage:
        """Publish a message to the broker.

        Send a MQTT [PUBLISH](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037)
        message and wait for acknowledgment depending on Quality Of Service

        Args:
            topic: topic name to which message data is published
            message: payload message (as bytes) to send.
            qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to `default_qos`
                config parameter or QOS_0.
            retain: retain flag. Defaults to ``default_retain`` config parameter or False.
            ack_timeout: duration to wait for connection acknowledgment from broker.

        Returns:
            the message that was sent

        """
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        def get_retain_and_qos() -> tuple[int, bool]:
            if qos is not None:
                if qos not in (QOS_0, QOS_1, QOS_2):
                    msg = f"QOS '{qos}' is not one of QOS_0, QOS_1, QOS_2."
                    raise ClientError(msg)
                _qos = qos
            else:
                _qos = self.config["default_qos"]
                with contextlib.suppress(KeyError):
                    _qos = self.config["topics"][topic]["qos"]
            if retain:
                _retain = retain
            else:
                _retain = self.config["default_retain"]
                with contextlib.suppress(KeyError):
                    _retain = self.config["topics"][topic]["retain"]
            return _qos, _retain

        (app_qos, app_retain) = get_retain_and_qos()
        return await self._handler.mqtt_publish(
            topic,
            message,
            app_qos,
            app_retain,
            ack_timeout,
        )

    @mqtt_connected
    async def subscribe(self, topics: list[tuple[str, int]]) -> list[int]:
        """Subscribe to topics.

        Send a MQTT [SUBSCRIBE](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718063)
        message and wait for broker acknowledgment.

        Args:
            topics: array of tuples containing topic pattern and QOS from `amqtt.mqtt.constants` to subscribe. For example:
                ```python
                [
                    ("$SYS/broker/uptime", QOS_1),
                    ("$SYS/broker/load/#", QOS_2),
                ]
                ```

        Returns:
            [SUBACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718068) message return code.

        """
        if self._handler and self.session:
            return await self._handler.mqtt_subscribe(topics, self.session.next_packet_id)
        return [0x80]

    @mqtt_connected
    async def unsubscribe(self, topics: list[str]) -> None:
        """Unsubscribe from topics.

        Send a MQTT [UNSUBSCRIBE](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718072)
        message and wait for broker [UNSUBACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718077)
        message.

        Args:
            topics: array of topics to unsubscribe from.
                ```
                ["$SYS/broker/uptime", "$SYS/broker/load/#"]
                ```

        """
        if self._handler and self.session:
            await self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    async def deliver_message(self, timeout_duration: float | None = None) -> ApplicationMessage | None:
        """Deliver the next received message.

        Deliver next message received from the broker. If no message is available, this methods waits until next message arrives
        or ``timeout_duration`` occurs.

        Args:
            timeout_duration: maximum number of seconds to wait before returning. If not specified or None, there is no limit.

        Returns:
            instance of `amqtt.session.ApplicationMessage` containing received message information flow.

        Raises:
            asyncio.TimeoutError: if timeout occurs before a message is delivered

        """
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        deliver_task = asyncio.create_task(self._handler.mqtt_deliver_next_message())
        self.client_tasks.append(deliver_task)
        self.logger.debug("Waiting for message delivery")

        done, _ = await asyncio.wait(
            [deliver_task],
            return_when=asyncio.FIRST_EXCEPTION,
            timeout=timeout_duration,
        )

        if self.client_tasks:
            self.client_tasks.pop()

        if deliver_task in done:
            exception = deliver_task.exception()
            if exception is not None:
                # deliver_task raised an exception, pass it on to our caller
                raise exception
            return deliver_task.result()
        # timeout occurred before message received
        deliver_task.cancel()
        msg = "Timeout waiting for message"
        raise asyncio.TimeoutError(msg)

    async def _connect_coro(self) -> int:
        """Perform the core connection logic."""
        if self.session is None:
            msg = "Session is not initialized."
            raise ClientError(msg)

        kwargs: dict[str, Any] = {}

        # Decode URI attributes
        uri_attributes = urlparse(self.session.broker_uri)
        scheme = uri_attributes.scheme
        secure = scheme in ("mqtts", "wss")
        self.session.username = (
            self.session.username
            if self.session.username
            else (str(uri_attributes.username) if uri_attributes.username else None)
        )
        self.session.password = (
            self.session.password
            if self.session.password
            else (str(uri_attributes.password) if uri_attributes.password else None)
        )
        self.session.remote_address = str(uri_attributes.hostname) if uri_attributes.hostname else None
        self.session.remote_port = uri_attributes.port

        if scheme in ("mqtt", "mqtts") and not self.session.remote_port:
            self.session.remote_port = 8883 if scheme == "mqtts" else 1883

        if scheme in ("ws", "wss") and not self.session.remote_port:
            self.session.remote_port = 443 if scheme == "wss" else 80

        if scheme in ("ws", "wss"):
            # Rewrite URI to conform to https://tools.ietf.org/html/rfc6455#section-3
            uri = (
                str(scheme),
                f"{self.session.remote_address}:{self.session.remote_port}",
                str(uri_attributes.path),
                str(uri_attributes.params),
                str(uri_attributes.query),
                str(uri_attributes.fragment),
            )
            self.session.broker_uri = str(urlunparse(uri))
        # Init protocol handler
        # if not self._handler:
        self._handler = ClientProtocolHandler(self.plugins_manager)

        connection_timeout = self.config.get("connection_timeout", None)

        if secure:
            sc = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH,
                cafile=self.session.cafile,
                capath=self.session.capath,
                cadata=self.session.cadata,
            )

            if "certfile" in self.config:
                sc.load_verify_locations(cafile=self.config["certfile"])
            if "check_hostname" in self.config and isinstance(self.config["check_hostname"], bool):
                sc.check_hostname = self.config["check_hostname"]
                sc.verify_mode = ssl.CERT_REQUIRED
            kwargs["ssl"] = sc

        try:
            reader: StreamReaderAdapter | WebSocketsReader | None = None
            writer: StreamWriterAdapter | WebSocketsWriter | None = None
            self._connected_state.clear()

            # Open connection
            if scheme in ("mqtt", "mqtts"):
                conn_reader, conn_writer = await asyncio.wait_for(
                    asyncio.open_connection(
                        self.session.remote_address,
                        self.session.remote_port,
                        **kwargs,
                    ), timeout=connection_timeout)

                reader = StreamReaderAdapter(conn_reader)
                writer = StreamWriterAdapter(conn_writer)
            elif scheme in ("ws", "wss") and self.session.broker_uri:
                websocket: ClientConnection = await asyncio.wait_for(
                    websockets.connect(
                        self.session.broker_uri,
                        subprotocols=[websockets.Subprotocol("mqtt")],
                        additional_headers=self.additional_headers,
                        **kwargs,
                    ), timeout=connection_timeout)

                reader = WebSocketsReader(websocket)
                writer = WebSocketsWriter(websocket)
            elif not self.session.broker_uri:
                msg = "missing broker uri"
                raise ClientError(msg)
            else:
                msg = f"incorrect scheme defined in uri: '{scheme!r}'"
                raise ClientError(msg)

            # Start MQTT protocol
            self._handler.attach(self.session, reader, writer)
            return_code: int | None = await self._handler.mqtt_connect()

            if return_code is not CONNECTION_ACCEPTED:
                self.session.transitions.disconnect()
                self.logger.warning(f"Connection rejected with code '{return_code}'")
                msg = "Connection rejected by broker"
                exc = ConnectError(msg)
                exc.return_code = return_code
                raise exc
            # Handle MQTT protocol
            await self._handler.start()
            self.session.transitions.connect()
            self._connected_state.set()
            self.logger.debug(f"Connected to {self.session.remote_address}:{self.session.remote_port}")

        except (InvalidURI, InvalidHandshake, ProtocolHandlerError, ConnectionError, OSError) as e:
            self.logger.debug(f"Connection failed : {self.session.broker_uri} [{e!r}]")
            self.session.transitions.disconnect()
            raise ConnectError(e) from e
        return return_code

    async def handle_connection_close(self) -> None:
        """Handle disconnection from the broker."""
        if self.session is None:
            msg = "Session is not initialized."
            raise ClientError(msg)
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        def cancel_tasks() -> None:
            self._no_more_connections.set()
            while self.client_tasks:
                task = self.client_tasks.popleft()
                if not task.done():
                    task.cancel(msg="Connection closed.")

        self.logger.debug("Monitoring broker disconnection")
        # Wait for disconnection from broker (like connection lost)
        await self._handler.wait_disconnect()
        self.logger.warning("Disconnected from broker")

        # Block client API
        self._connected_state.clear()

        # stop an clean handler
        await self._handler.stop()
        self._handler.detach()
        self.session.transitions.disconnect()

        if self.config.get("auto_reconnect", False):
            # Try reconnection
            self.logger.debug("Auto-reconnecting")
            try:
                await self.reconnect()
            except ConnectError:
                # Cancel client pending tasks
                cancel_tasks()
        else:
            # Cancel client pending tasks
            cancel_tasks()

    def _init_session(
        self,
        uri: str | None = None,
        cleansession: bool | None = None,
        cafile: str | None = None,
        capath: str | None = None,
        cadata: str | None = None,
    ) -> Session:
        """Initialize the MQTT session."""
        broker_conf = self.config.get("broker", {}).copy()
        broker_conf.update(
            {k: v for k, v in {"uri": uri, "cafile": cafile, "capath": capath, "cadata": cadata}.items() if v is not None},
        )

        if not broker_conf.get("uri"):
            msg = "Missing connection parameter 'uri'"
            raise ClientError(msg)

        session = Session()
        session.broker_uri = broker_conf["uri"]
        session.client_id = self.client_id
        session.cafile = broker_conf.get("cafile")
        session.capath = broker_conf.get("capath")
        session.cadata = broker_conf.get("cadata")

        if cleansession is not None:
            broker_conf["cleansession"] = cleansession  # noop?
            session.clean_session = cleansession
        else:
            session.clean_session = self.config.get("cleansession", True)

        session.keep_alive = self.config["keep_alive"] - self.config["ping_delay"]

        if "will" in self.config:
            session.will_flag = True
            session.will_retain = self.config["will"]["retain"]
            session.will_topic = self.config["will"]["topic"]
            session.will_message = self.config["will"]["message"].encode()
            session.will_qos = self.config["will"]["qos"]

        return session



from abc import ABC, abstractmethod
from asyncio import StreamReader, StreamWriter
from contextlib import suppress
import io
import logging

from websockets import ConnectionClosed
from websockets.asyncio.connection import Connection


class ReaderAdapter(ABC):
    """Base class for all network protocol reader adapters.

    Reader adapters are used to adapt read operations on the network depending on the
    protocol used.
    """

    @abstractmethod
    async def read(self, n: int = -1) -> bytes:
        """Read up to n bytes. If n is not provided, or set to -1, read until EOF and return all read bytes.

        If the EOF was received and the internal buffer is
        empty, return an empty bytes object. :return: packet read as bytes data.
        """
        raise NotImplementedError

    @abstractmethod
    def feed_eof(self) -> None:
        """Acknowledge EOF."""
        raise NotImplementedError


class WriterAdapter(ABC):
    """Base class for all network protocol writer adapters.

    Writer adapters are used to adapt write operations on the network depending on
    the protocol used.
    """

    @abstractmethod
    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        raise NotImplementedError

    @abstractmethod
    async def drain(self) -> None:
        """Let the write buffer of the underlying transport a chance to be flushed."""
        raise NotImplementedError

    @abstractmethod
    def get_peer_info(self) -> tuple[str, int] | None:
        """Return peer socket info (remote address and remote port as tuple)."""
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Close the protocol connection."""
        raise NotImplementedError


class WebSocketsReader(ReaderAdapter):
    """WebSockets API reader adapter.

    This adapter relies on Connection to read from a WebSocket.
    """

    def __init__(self, protocol: Connection) -> None:
        self._protocol = protocol
        self._stream = io.BytesIO(b"")

    async def read(self, n: int = -1) -> bytes:
        await self._feed_buffer(n)
        return self._stream.read(n)

    async def _feed_buffer(self, n: int = 1) -> None:
        """Feed the data buffer by reading a WebSocket message.

        :param n: Optional; feed buffer until it contains at least n bytes. Defaults to 1.
        """
        buffer = bytearray(self._stream.read())
        message: str | bytes | None = None
        while len(buffer) < n:
            with suppress(ConnectionClosed):
                message = await self._protocol.recv()
            if message is None:
                break
            message = message.encode("utf-8") if isinstance(message, str) else message
            buffer.extend(message)
        self._stream = io.BytesIO(buffer)

    def feed_eof(self) -> None:
        # NOTE: not implemented?!
        pass


class WebSocketsWriter(WriterAdapter):
    """WebSockets API writer adapter.

    This adapter relies on Connection to write to a WebSocket.
    """

    def __init__(self, protocol: Connection) -> None:
        self._protocol = protocol
        self._stream = io.BytesIO(b"")

    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        self._stream.write(data)

    async def drain(self) -> None:
        """Let the write buffer of the underlying transport a chance to be flushed."""
        data = self._stream.getvalue()
        if data and len(data):
            await self._protocol.send(data)
        self._stream = io.BytesIO(b"")

    def get_peer_info(self) -> tuple[str, int] | None:
        # remote_address can be either a 4-tuple or 2-tuple depending on whether
        # it is an IPv6 or IPv4 address, so we take their shared (host, port)
        # prefix here to present a uniform return value.
        remote_address: tuple[str, int] | None = self._protocol.remote_address[:2]
        return remote_address

    async def close(self) -> None:
        await self._protocol.close()


class StreamReaderAdapter(ReaderAdapter):
    """Asyncio Streams API protocol adapter.

    This adapter relies on StreamReader to read from a TCP socket.
    Because API is very close, this class is trivial.
    """

    def __init__(self, reader: StreamReader) -> None:
        self._reader = reader

    async def read(self, n: int = -1) -> bytes:
        if n == -1:
            data = await self._reader.read(n)
        else:
            data = await self._reader.readexactly(n)
        return data

    def feed_eof(self) -> None:
        self._reader.feed_eof()


class StreamWriterAdapter(WriterAdapter):
    """Asyncio Streams API protocol adapter.

    This adapter relies on StreamWriter to write to a TCP socket.
    Because API is very close, this class is trivial.
    """

    def __init__(self, writer: StreamWriter) -> None:
        self.logger = logging.getLogger(__name__)
        self._writer = writer
        self.is_closed = False  # StreamWriter has no test for closed...we use our own

    def write(self, data: bytes) -> None:
        if not self.is_closed:
            self._writer.write(data)

    async def drain(self) -> None:
        if not self.is_closed:
            await self._writer.drain()

    def get_peer_info(self) -> tuple[str, int]:
        extra_info = self._writer.get_extra_info("peername")
        return extra_info[0], extra_info[1]

    async def close(self) -> None:
        if not self.is_closed:
            self.is_closed = True  # we first mark this closed so yields below don't cause races with waiting writes
            await self._writer.drain()
            if self._writer.can_write_eof():
                self._writer.write_eof()
            self._writer.close()
            with suppress(AttributeError):
                await self._writer.wait_closed()


class BufferReader(ReaderAdapter):
    """Byte Buffer reader adapter.

    This adapter simply adapts reading a byte buffer.
    """

    def __init__(self, buffer: bytes) -> None:
        self._stream = io.BytesIO(buffer)

    async def read(self, n: int = -1) -> bytes:
        return self._stream.read(n)

    def feed_eof(self) -> None:
        # NOTE: not implemented?!
        pass


class BufferWriter(WriterAdapter):
    """ByteBuffer writer adapter.

    This adapter simply adapts writing to a byte buffer.
    """

    def __init__(self, buffer: bytes = b"") -> None:
        self._stream = io.BytesIO(buffer)

    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        self._stream.write(data)

    async def drain(self) -> None:
        pass

    def get_buffer(self) -> bytes:
        return self._stream.getvalue()

    def get_peer_info(self) -> tuple[str, int]:
        return "BufferWriter", 0

    async def close(self) -> None:
        self._stream.close()



import asyncio
from typing import TYPE_CHECKING, Any

from amqtt.errors import AMQTTError, NoDataError
from amqtt.events import MQTTEvents
from amqtt.mqtt.connack import ConnackPacket
from amqtt.mqtt.connect import ConnectPacket, ConnectPayload, ConnectVariableHeader
from amqtt.mqtt.disconnect import DisconnectPacket
from amqtt.mqtt.pingreq import PingReqPacket
from amqtt.mqtt.pingresp import PingRespPacket
from amqtt.mqtt.protocol.handler import ProtocolHandler
from amqtt.mqtt.suback import SubackPacket
from amqtt.mqtt.subscribe import SubscribePacket
from amqtt.mqtt.unsuback import UnsubackPacket
from amqtt.mqtt.unsubscribe import UnsubscribePacket
from amqtt.plugins.manager import PluginManager
from amqtt.session import Session

if TYPE_CHECKING:
    from amqtt.client import ClientContext

class ClientProtocolHandler(ProtocolHandler["ClientContext"]):
    def __init__(
        self,
        plugins_manager: PluginManager["ClientContext"],
        session: Session | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        super().__init__(plugins_manager, session, loop=loop)
        self._ping_task: asyncio.Task[Any] | None = None
        self._pingresp_queue: asyncio.Queue[PingRespPacket] = asyncio.Queue()
        self._subscriptions_waiter: dict[int, asyncio.Future[list[int]]] = {}
        self._unsubscriptions_waiter: dict[int, asyncio.Future[Any]] = {}
        self._disconnect_waiter: asyncio.Future[Any] | None = asyncio.Future()

    async def start(self) -> None:
        await super().start()
        if self._disconnect_waiter and self._disconnect_waiter.cancelled():
            self._disconnect_waiter = asyncio.Future()

    async def stop(self) -> None:
        await super().stop()
        if self._ping_task and not self._ping_task.cancelled():
            self.logger.debug("Cancel ping task")
            self._ping_task.cancel()

        if self._disconnect_waiter and not self._disconnect_waiter.done():
            self._disconnect_waiter.cancel()

    def _build_connect_packet(self) -> ConnectPacket:
        vh = ConnectVariableHeader()
        payload = ConnectPayload()

        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)

        vh.keep_alive = self.session.keep_alive
        vh.clean_session_flag = self.session.clean_session if self.session.clean_session is not None else False
        vh.will_retain_flag = self.session.will_retain if self.session.will_retain is not None else False
        payload.client_id = self.session.client_id

        if self.session.username:
            vh.username_flag = True
            payload.username = self.session.username
        else:
            vh.username_flag = False

        if self.session.password:
            vh.password_flag = True
            payload.password = self.session.password
        else:
            vh.password_flag = False

        if self.session.will_flag:
            vh.will_flag = True
            if self.session.will_qos is not None:
                vh.will_qos = self.session.will_qos
            payload.will_message = self.session.will_message
            payload.will_topic = self.session.will_topic
        else:
            vh.will_flag = False

        return ConnectPacket(variable_header=vh, payload=payload)

    async def mqtt_connect(self) -> int | None:
        connect_packet = self._build_connect_packet()
        await self._send_packet(connect_packet)

        if self.reader is None:
            msg = "Reader is not initialized."
            raise AMQTTError(msg)
        try:
            connack = await ConnackPacket.from_stream(self.reader)
        except NoDataError as e:
            raise ConnectionError from e
        await self.plugins_manager.fire_event(MQTTEvents.PACKET_RECEIVED, packet=connack, session=self.session)
        return connack.return_code

    def handle_write_timeout(self) -> None:
        try:
            if not self._ping_task:
                self.logger.debug("Scheduling Ping")
                self._ping_task = asyncio.create_task(self.mqtt_ping())
        except asyncio.InvalidStateError as e:
            self.logger.warning(f"Invalid state while scheduling ping task: {e!r}")
        except asyncio.CancelledError as e:
            self.logger.info(f"Ping task was cancelled: {e!r}")

    def handle_read_timeout(self) -> None:
        pass

    async def mqtt_subscribe(self, topics: list[tuple[str, int]], packet_id: int) -> list[int]:
        """Subscribe to the given topics.

        :param topics: List of tuples, e.g. [('filter', '/a/b', 'qos': 0x00)].
        :return: Return codes for the subscription.
        """
        subscribe = SubscribePacket.build(topics, packet_id)
        await self._send_packet(subscribe)

        if subscribe.variable_header is None:
            msg = f"Invalid variable header in SUBSCRIBE packet: {subscribe.variable_header}"
            raise AMQTTError(msg)

        waiter: asyncio.Future[list[int]] = asyncio.Future()
        self._subscriptions_waiter[subscribe.variable_header.packet_id] = waiter
        try:
            return_codes = await waiter
        finally:
            del self._subscriptions_waiter[subscribe.variable_header.packet_id]
        return return_codes

    async def handle_suback(self, suback: SubackPacket) -> None:
        if suback.variable_header is None:
            msg = "SUBACK packet: variable header not initialized."
            raise AMQTTError(msg)
        if suback.payload is None:
            msg = "SUBACK packet: payload not initialized."
            raise AMQTTError(msg)

        packet_id = suback.variable_header.packet_id

        waiter = self._subscriptions_waiter.get(packet_id)
        if waiter is not None:
            waiter.set_result(suback.payload.return_codes)
        else:
            self.logger.warning(f"Received SUBACK for unknown pending subscription with Id: {packet_id}")

    async def mqtt_unsubscribe(self, topics: list[str], packet_id: int) -> None:
        """Unsubscribe from the given topics.

        :param topics: List of topics ['/a/b', ...].
        """
        unsubscribe = UnsubscribePacket.build(topics, packet_id)

        if unsubscribe.variable_header is None:
            msg = "UNSUBSCRIBE packet: variable header not initialized."
            raise AMQTTError(msg)

        await self._send_packet(unsubscribe)
        waiter: asyncio.Future[Any] = asyncio.Future()
        self._unsubscriptions_waiter[unsubscribe.variable_header.packet_id] = waiter
        try:
            await waiter
        finally:
            del self._unsubscriptions_waiter[unsubscribe.variable_header.packet_id]

    async def handle_unsuback(self, unsuback: UnsubackPacket) -> None:
        if unsuback.variable_header is None:
            msg = "UNSUBACK packet: variable header not initialized."
            raise AMQTTError(msg)

        packet_id = unsuback.variable_header.packet_id
        waiter = self._unsubscriptions_waiter.get(packet_id)
        if waiter is not None:
            waiter.set_result(None)
        else:
            self.logger.warning(f"Received UNSUBACK for unknown pending unsubscription with Id: {packet_id}")

    async def mqtt_disconnect(self) -> None:
        disconnect_packet = DisconnectPacket()
        await self._send_packet(disconnect_packet)

    async def mqtt_ping(self) -> PingRespPacket:
        ping_packet = PingReqPacket()
        try:
            await self._send_packet(ping_packet)
            resp = await self._pingresp_queue.get()
        finally:
            self._ping_task = None  # Ensure the task is cleaned up
        return resp

    async def handle_pingresp(self, pingresp: PingRespPacket) -> None:
        await self._pingresp_queue.put(pingresp)

    async def handle_connection_closed(self) -> None:
        self.logger.debug("Broker closed connection")
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def wait_disconnect(self) -> None:
        if self._disconnect_waiter is not None:
            await self._disconnect_waiter



from asyncio import StreamReader
from typing_extensions import Self

from amqtt.adapters import ReaderAdapter
from amqtt.codecs_amqtt import decode_string, encode_string
from amqtt.errors import AMQTTError, NoDataError
from amqtt.mqtt.packet import UNSUBSCRIBE, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader, PacketIdVariableHeader


class UnubscribePayload(MQTTPayload[MQTTVariableHeader]):
    __slots__ = ("topics",)

    def __init__(self, topics: list[str] | None = None) -> None:
        super().__init__()
        self.topics = topics or []

    def to_bytes(self, fixed_header: MQTTFixedHeader | None = None, variable_header: MQTTVariableHeader | None = None) -> bytes:
        out = b""
        for topic in self.topics:
            out += encode_string(topic)
        return out

    @classmethod
    async def from_stream(
        cls: type[Self],
        reader: StreamReader | ReaderAdapter,
        fixed_header: MQTTFixedHeader | None,
        variable_header: MQTTVariableHeader | None,
    ) -> Self:
        if fixed_header is None or variable_header is None:
            msg = "Fixed header or Value header is not set."
            raise ValueError(msg)

        topics = []
        payload_length = fixed_header.remaining_length - variable_header.bytes_length
        read_bytes = 0
        while read_bytes < payload_length:
            try:
                topic = await decode_string(reader)
                topics.append(topic)
                read_bytes += 2 + len(topic.encode("utf-8"))
            except NoDataError:
                break
        return cls(topics)


class UnsubscribePacket(MQTTPacket[PacketIdVariableHeader, UnubscribePayload, MQTTFixedHeader]):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = UnubscribePayload

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: PacketIdVariableHeader | None = None,
        payload: UnubscribePayload | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(UNSUBSCRIBE, 0x02)  # [MQTT-3.10.1-1]
        else:
            if fixed.packet_type is not UNSUBSCRIBE:
                msg = f"Invalid fixed packet type {fixed.packet_type} for UnsubscribePacket init"
                raise AMQTTError(msg)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topics: list[str], packet_id: int) -> "UnsubscribePacket":
        v_header = PacketIdVariableHeader(packet_id)
        payload = UnubscribePayload(topics)
        return UnsubscribePacket(variable_header=v_header, payload=payload)



try:
    from enum import StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  #type: ignore[no-redef]
        pass


class Events(StrEnum):
    """Class for all events."""


class ClientEvents(Events):
    """Events issued by the client."""


class MQTTEvents(Events):
    PACKET_SENT = "mqtt_packet_sent"
    PACKET_RECEIVED = "mqtt_packet_received"


class BrokerEvents(Events):
    """Events issued by the broker."""

    PRE_START = "broker_pre_start"
    POST_START = "broker_post_start"
    PRE_SHUTDOWN = "broker_pre_shutdown"
    POST_SHUTDOWN = "broker_post_shutdown"
    CLIENT_CONNECTED = "broker_client_connected"
    CLIENT_DISCONNECTED = "broker_client_disconnected"
    CLIENT_SUBSCRIBED = "broker_client_subscribed"
    CLIENT_UNSUBSCRIBED = "broker_client_unsubscribed"
    MESSAGE_RECEIVED = "broker_message_received"



import asyncio
from asyncio import AbstractEventLoop, Queue
from typing import TYPE_CHECKING

from amqtt.adapters import ReaderAdapter, WriterAdapter
from amqtt.errors import MQTTError
from amqtt.events import MQTTEvents
from amqtt.mqtt.connack import (
    BAD_USERNAME_PASSWORD,
    CONNECTION_ACCEPTED,
    IDENTIFIER_REJECTED,
    NOT_AUTHORIZED,
    UNACCEPTABLE_PROTOCOL_VERSION,
    ConnackPacket,
)
from amqtt.mqtt.connect import ConnectPacket
from amqtt.mqtt.disconnect import DisconnectPacket
from amqtt.mqtt.pingreq import PingReqPacket
from amqtt.mqtt.pingresp import PingRespPacket
from amqtt.mqtt.protocol.handler import ProtocolHandler
from amqtt.mqtt.suback import SubackPacket
from amqtt.mqtt.subscribe import SubscribePacket
from amqtt.mqtt.unsuback import UnsubackPacket
from amqtt.mqtt.unsubscribe import UnsubscribePacket
from amqtt.plugins.manager import PluginManager
from amqtt.session import Session
from amqtt.utils import format_client_message

_MQTT_PROTOCOL_LEVEL_SUPPORTED = 4

if TYPE_CHECKING:
    from amqtt.broker import BrokerContext

class Subscription:
    def __init__(self, packet_id: int, topics: list[tuple[str, int]]) -> None:
        self.packet_id = packet_id
        self.topics = topics


class UnSubscription:
    def __init__(self, packet_id: int, topics: list[str]) -> None:
        self.packet_id = packet_id
        self.topics = topics


class BrokerProtocolHandler(ProtocolHandler["BrokerContext"]):
    def __init__(
        self,
        plugins_manager: PluginManager["BrokerContext"],
        session: Session | None = None,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        super().__init__(plugins_manager, session, loop)
        self._disconnect_waiter: asyncio.Future[DisconnectPacket | None] | None = None
        self._pending_subscriptions: Queue[Subscription] = Queue()
        self._pending_unsubscriptions: Queue[UnSubscription] = Queue()

    async def start(self) -> None:
        await super().start()
        # Ensure the disconnect waiter is reset
        if self._disconnect_waiter is None or self._disconnect_waiter.done():
            self._disconnect_waiter = asyncio.Future()

    async def stop(self) -> None:
        """Stop the protocol handler and reset the disconnect waiter."""
        await super().stop()
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)
        self._disconnect_waiter = None  # Reset the disconnect waiter
        # Clear pending subscriptions and unsubscriptions
        while not self._pending_subscriptions.empty():
            self._pending_subscriptions.get_nowait()
        while not self._pending_unsubscriptions.empty():
            self._pending_unsubscriptions.get_nowait()

    async def wait_disconnect(self) -> DisconnectPacket | None:
        """Wait for a disconnect packet or connection closure."""
        if self._disconnect_waiter is not None:
            return await self._disconnect_waiter
        return None

    def handle_write_timeout(self) -> None:
        pass

    def handle_read_timeout(self) -> None:
        pass

    async def handle_disconnect(self, disconnect: DisconnectPacket | None) -> None:
        """Handle a disconnect packet and notify the disconnect waiter."""
        self.logger.debug("Client disconnecting")
        if self._disconnect_waiter and not self._disconnect_waiter.done():
            self.logger.debug(f"Setting disconnect waiter result to {disconnect!r}")
            self._disconnect_waiter.set_result(disconnect)
        self._disconnect_waiter = None  # Reset the disconnect waiter to avoid reuse

    async def handle_connection_closed(self) -> None:
        """Handle connection closure and notify the disconnect waiter."""
        await self.handle_disconnect(None)

    async def handle_connect(self, connect: ConnectPacket) -> None:
        # Broker handler shouldn't receive CONNECT message during messages handling
        # as CONNECT messages are managed by the broker on client connection
        self.logger.error(
            f"{self.session.client_id if self.session else None} [MQTT-3.1.0-2] {format_client_message(self.session)} :"
            f" CONNECT message received during messages handling",
        )
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def handle_pingreq(self, pingreq: PingReqPacket) -> None:
        await self._send_packet(PingRespPacket.build())

    async def handle_subscribe(self, subscribe: SubscribePacket) -> None:
        if subscribe.variable_header is None:
            msg = "SUBSCRIBE packet: variable header not initialized."
            raise MQTTError(msg)
        if subscribe.payload is None:
            msg = "SUBSCRIBE packet: payload not initialized."
            raise MQTTError(msg)

        subscription: Subscription = Subscription(subscribe.variable_header.packet_id, subscribe.payload.topics)
        await self._pending_subscriptions.put(subscription)

    async def handle_unsubscribe(self, unsubscribe: UnsubscribePacket) -> None:
        if unsubscribe.variable_header is None:
            msg = "UNSUBSCRIBE packet: variable header not initialized."
            raise MQTTError(msg)
        if unsubscribe.payload is None:
            msg = "UNSUBSCRIBE packet: payload not initialized."
            raise MQTTError(msg)
        unsubscription: UnSubscription = UnSubscription(unsubscribe.variable_header.packet_id, unsubscribe.payload.topics)
        await self._pending_unsubscriptions.put(unsubscription)

    async def get_next_pending_subscription(self) -> Subscription:
        return await self._pending_subscriptions.get()

    async def get_next_pending_unsubscription(self) -> UnSubscription:
        return await self._pending_unsubscriptions.get()

    async def mqtt_acknowledge_subscription(self, packet_id: int, return_codes: list[int]) -> None:
        suback = SubackPacket.build(packet_id, return_codes)
        await self._send_packet(suback)

    async def mqtt_acknowledge_unsubscription(self, packet_id: int) -> None:
        unsuback = UnsubackPacket.build(packet_id)
        await self._send_packet(unsuback)

    async def mqtt_connack_authorize(self, authorize: bool) -> None:
        if self.session is None:
            msg = "Session is not initialized!"
            raise MQTTError(msg)

        connack = ConnackPacket.build(self.session.parent, CONNECTION_ACCEPTED if authorize else NOT_AUTHORIZED)
        await self._send_packet(connack)

    @classmethod
    async def init_from_connect(
        cls,
        reader: ReaderAdapter,
        writer: WriterAdapter,
        plugins_manager: PluginManager["BrokerContext"],
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> tuple["BrokerProtocolHandler", Session]:
        """Initialize from a CONNECT packet and validates the connection."""
        connect = await ConnectPacket.from_stream(reader)
        await plugins_manager.fire_event(MQTTEvents.PACKET_RECEIVED, packet=connect)

        if connect.variable_header is None:
            msg = "CONNECT packet: variable header not initialized."
            raise MQTTError(msg)
        if connect.payload is None:
            msg = "CONNECT packet: payload not initialized."
            raise MQTTError(msg)

        # this shouldn't be required anymore since broker generates for each client a random client_id if not provided
        # [MQTT-3.1.3-6]
        if connect.payload.client_id is None:
            msg = "[[MQTT-3.1.3-3]] : Client identifier must be present"
            raise MQTTError(msg)

        if connect.variable_header.will_flag and (connect.payload.will_topic is None or connect.payload.will_message is None):
            msg = "Will flag set, but will topic/message not present in payload"
            raise MQTTError(msg)

        if connect.variable_header.reserved_flag:
            msg = "[MQTT-3.1.2-3] CONNECT reserved flag must be set to 0"
            raise MQTTError(msg)

        if connect.proto_name != "MQTT":
            msg = f'[MQTT-3.1.2-1] Incorrect protocol name: "{connect.proto_name}"'
            raise MQTTError(msg)

        remote_info = writer.get_peer_info()
        if remote_info is not None:
            remote_address, remote_port = remote_info
            connack = None
            error_msg = None
            if connect.proto_level != _MQTT_PROTOCOL_LEVEL_SUPPORTED:
                # only MQTT 3.1.1 supported
                error_msg = (
                    f"Invalid protocol from {format_client_message(address=remote_address, port=remote_port)}:"
                    f" {connect.proto_level}"
                )
                connack = ConnackPacket.build(0, UNACCEPTABLE_PROTOCOL_VERSION)  # [MQTT-3.2.2-4] session_parent=0
            elif not connect.username_flag and connect.password_flag:
                connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.1.2-22]
            elif connect.username_flag and connect.username is None:
                error_msg = f"Invalid username from {format_client_message(address=remote_address, port=remote_port)}"
                connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.2.2-4] session_parent=0
            elif connect.password_flag and connect.password is None:
                error_msg = f"Invalid password from {format_client_message(address=remote_address, port=remote_port)}"
                connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.2.2-4] session_parent=0
            elif connect.clean_session_flag is False and connect.payload.client_id_is_random:
                error_msg = (
                    f"[MQTT-3.1.3-8] [MQTT-3.1.3-9] {format_client_message(address=remote_address, port=remote_port)}:"
                    " No client Id provided (cleansession=0)"
                )
                connack = ConnackPacket.build(0, IDENTIFIER_REJECTED)

            if connack is not None:
                await plugins_manager.fire_event(MQTTEvents.PACKET_SENT, packet=connack)
                await connack.to_stream(writer)
                await writer.close()
                raise MQTTError(error_msg) from None

        incoming_session = Session()
        incoming_session.client_id = connect.client_id
        incoming_session.clean_session = connect.clean_session_flag
        incoming_session.will_flag = connect.will_flag
        incoming_session.will_retain = connect.will_retain_flag
        incoming_session.will_qos = connect.will_qos
        incoming_session.will_topic = connect.will_topic
        incoming_session.will_message = connect.will_message
        incoming_session.username = connect.username
        incoming_session.password = connect.password
        incoming_session.remote_address = remote_address
        incoming_session.remote_port = remote_port

        incoming_session.keep_alive = max(connect.keep_alive, 0)

        if connect.keep_alive > 0:
            incoming_session.keep_alive = connect.keep_alive
        else:
            incoming_session.keep_alive = 0

        handler = cls(plugins_manager, loop=loop)
        return handler, incoming_session



from collections.abc import Callable, Coroutine
from functools import partial
import logging
from typing import Any, TypeAlias

from amqtt.contexts import BaseContext
from amqtt.events import BrokerEvents
from amqtt.mqtt import MQTTPacket
from amqtt.mqtt.packet import MQTTFixedHeader, MQTTPayload, MQTTVariableHeader
from amqtt.plugins.base import BasePlugin
from amqtt.session import Session

PACKET: TypeAlias = MQTTPacket[MQTTVariableHeader, MQTTPayload[MQTTVariableHeader], MQTTFixedHeader]


class EventLoggerPlugin(BasePlugin[BaseContext]):
    """A plugin to log events dynamically based on method names."""

    async def log_event(self, *args: Any, **kwargs: Any) -> None:
        """Log the occurrence of an event."""
        event_name = kwargs["event_name"].replace("old", "")
        if event_name.replace("on_", "") in (BrokerEvents.CLIENT_CONNECTED, BrokerEvents.CLIENT_DISCONNECTED):
            self.context.logger.info(f"### '{event_name}' EVENT FIRED ###")
        else:
            self.context.logger.debug(f"### '{event_name}' EVENT FIRED ###")

    def __getattr__(self, name: str) -> Callable[..., Coroutine[Any, Any, None]]:
        """Dynamically handle calls to methods starting with 'on_'."""
        if name.startswith("on_"):
            return partial(self.log_event, event_name=name)
        msg = f"'EventLoggerPlugin' object has no attribute {name!r}"
        raise AttributeError(msg)


class PacketLoggerPlugin(BasePlugin[BaseContext]):
    """A plugin to log MQTT packets sent and received."""

    async def on_mqtt_packet_received(self, *, packet: PACKET, session: Session | None = None) -> None:
        """Log an MQTT packet when it is received."""
        if self.context.logger.isEnabledFor(logging.DEBUG):
            if session is not None:
                self.context.logger.debug(f"{session.client_id} <-in-- {packet!r}")
            else:
                self.context.logger.debug(f"<-in-- {packet!r}")

    async def on_mqtt_packet_sent(self, *, packet: PACKET, session: Session | None = None) -> None:
        """Log an MQTT packet when it is sent."""
        if self.context.logger.isEnabledFor(logging.DEBUG):
            if session is not None:
                self.context.logger.debug(f"{session.client_id} -out-> {packet!r}")
            else:
                self.context.logger.debug(f"-out-> {packet!r}")
